import logging
from warnings import warn

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.HardwareObjects.abstract.AbstractMotor import (
    AbstractMotor,
    MotorStates,
)


class MD2TimeoutError(Exception):
    pass


class MD2Motor(AbstractMotor):
    def __init__(self, name):
        AbstractMotor.__init__(self, name)
        self.motor_pos_attr_suffix = "Position"

    def init(self):
        self.motor_state = MotorStates.UNKNOWN
        if self.actuator_name in [None, ""]:
            self.actuator_name = self.get_property("actuator_name")

        self.motor_resolution = self.get_property("resolution")
        if self.motor_resolution is None:
            self.motor_resolution = 1e-3

        self.position_attr = self.add_channel(
            {"type": "exporter", "name": "position"},
            self.actuator_name + self.motor_pos_attr_suffix,
        )

        if self.position_attr is not None:
            self.position_attr.connect_signal("update", self.update_value)

            self.state_attr = self.add_channel(
                {"type": "exporter", "name": "%sState" % self.actuator_name}, "State"
            )

            self.motors_state_attr = self.add_channel(
                {"type": "exporter", "name": "MotorStates"}, "MotorStates"
            )
            if self.motors_state_attr is not None:
                self.motors_state_attr.connect_signal("update", self.updateMotorState)

            self._motor_abort = self.add_command(
                {"type": "exporter", "name": "abort"}, "abort"
            )

            self.get_limits_cmd = self.add_command(
                {"type": "exporter", "name": "get%sLimits" % self.actuator_name},
                "getMotorLimits",
            )
            self.get_dynamic_limits_cmd = self.add_command(
                {"type": "exporter", "name": "get%sDynamicLimits" % self.actuator_name},
                "getMotorDynamicLimits",
            )

            self.home_cmd = self.add_command(
                {"type": "exporter", "name": "%sHoming" % self.actuator_name},
                "startHomingMotor",
            )

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emit("valueChanged", (self.get_value(),))
        elif signal == "stateChanged":
            self.updateMotorState(self.motors_state_attr.get_value())
        elif signal == "limitsChanged":
            self.motorLimitsChanged()

    def updateMotorState(self, motor_states):
        d = dict([x.split("=") for x in motor_states])

        # new_motor_state = MotorStates.DESC_TO_STATE[d[self.actuator_name]]
        new_motor_state = MotorStates.__members__[d[self.actuator_name].upper()]

        if self.motor_state == new_motor_state:
            return

        self.motor_state = new_motor_state
        self.updateState()
        self.motorStateChanged(new_motor_state)

    def motorStateChanged(self, state):
        logging.getLogger().debug(
            "{}: in motorStateChanged: motor state changed to {}".format(
                self.name(), state
            )
        )
        self.emit("stateChanged", (state,))

    # Replaced by AbstractMotor.update_value:
    #
    # NB - was already broken (__position not set)
    #
    # def motor_positions_changed(self, position, private={}):
    #     """
    #     logging.getLogger().debug(
    #         "{}: in motor_positions_changed: motor position changed to {}".format(self.name(), position))
    #     """
    #     if abs(position - self.__position) <= self.motor_resolution:
    #         return
    #     self.__position = position
    #     print("%s --- %s" % (position, self.__position))
    #     self.emit("valueChanged", (self.__position,))

    def motorLimitsChanged(self):
        self.emit("limitsChanged", (self.get_limits(),))

    def get_state(self):
        return self.motor_state

    def get_dynamic_limits(self):
        try:
            low_lim, hi_lim = map(
                float, self.get_dynamic_limits_cmd(self.actuator_name)
            )
            if low_lim == float(1e999) or hi_lim == float(1e999):
                raise ValueError
            return low_lim, hi_lim
        except Exception:
            return (-1e4, 1e4)

    def get_limits(self):
        try:
            low_lim, hi_lim = map(float, self.get_limits_cmd(self.actuator_name))
            if low_lim == float(1e999) or hi_lim == float(1e999):
                raise ValueError
            return low_lim, hi_lim
        except Exception:
            return (-1e4, 1e4)

    def get_value(self):
        ret = self.position_attr.get_value()
        if ret is None:
            raise RuntimeError("%s: motor position is None" % self.name())
        return ret

    def _set_value(self, value):
        self.position_attr.set_value(value)

    def waitEndOfMove(self, timeout=None):
        self.wait_end_of_move(timeout)

    def wait_end_of_move(self, timeout=None):
        with Timeout(timeout):
            sleep(0.1)
            while self.motor_state == MotorStates.MOVING:
                sleep(0.1)

    def motorIsMoving(self):
        warn("motorIsMoving is deprecated. Use is_ready instead", DeprecationWarning)
        return self.is_ready() and self.motor_state == MotorStates.MOVING

    def get_motor_mnemonic(self):
        return self.actuator_name

    def stop(self):
        if self.get_state() != MotorStates.NOTINITIALIZED:
            self._motor_abort()

    def home_motor(self, timeout=None):
        self.home_cmd(self.actuator_name)
        try:
            self.waitEndOfMove(timeout)
        except Exception:
            raise MD2TimeoutError



import logging

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.CommandContainer import CommandObject
from mxcubecore.TaskUtils import task


class ControllerCommand(CommandObject):
    def __init__(self, name, cmd):
        CommandObject.__init__(self, name)
        self._cmd = cmd
        self._cmd_execution = None

    def is_connected(self):
        return True

    def get_arguments(self):
        if self.name() == "Anneal":
            self._arguments.append(("Time [s]", "float"))
        return self._arguments

    @task
    def __call__(self, *args, **kwargs):
        logging.getLogger("user_level_log").info("Starting %s" % self.name())
        self.emit("commandBeginWaitReply", (str(self.name()),))
        self._cmd_execution = gevent.spawn(self._cmd, *args, **kwargs)
        self._cmd_execution.link(self._cmd_done)

    def _cmd_done(self, cmd_execution):
        try:
            try:
                res = cmd_execution.get()
            except Exception:
                logging.getLogger("user_level_log").exception(
                    str(self.name()) + " failed!"
                )
                self.emit("commandFailed", (str(self.name()),))
            else:
                logging.getLogger("user_level_log").info(
                    str(self.name()) + " finished successfully."
                )
                if isinstance(res, gevent.GreenletExit):
                    self.emit("commandFailed", (str(self.name()),))
                else:
                    self.emit("commandReplyArrived", (str(self.name()), res))
        finally:
            self.emit("commandReady")

    def abort(self):
        if self._cmd_execution and not self._cmd_execution.ready():
            self._cmd_execution.kill()


class ID30BeamCmds(HardwareObject):
    def __init__(self, *args):
        HardwareObject.__init__(self, *args)

    def init(self):
        controller = self.get_object_by_role("controller")
        self.centrebeam = ControllerCommand("Centre beam", controller.centrebeam)
        self.quick_realign = ControllerCommand(
            "Quick realign", controller.quick_realign
        )
        self.anneal = ControllerCommand("Anneal", controller.anneal_procedure)

    def get_commands(self):
        return [self.centrebeam, self.quick_realign, self.anneal]



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Mockup class for testing purposes"""

import time

from mxcubecore.HardwareObjects.abstract.AbstractEnergy import AbstractEnergy
from mxcubecore.HardwareObjects.mockup.ActuatorMockup import ActuatorMockup

# Default energy value (keV)
DEFAULT_VALUE = 12.4
# Default energy limits (keV)
DEFAULT_LIMITS = (4, 20)


class EnergyMockup(ActuatorMockup, AbstractEnergy):
    """Energy Mockup class"""

    def init(self):
        """Initialise default properties"""
        super(EnergyMockup, self).init()

        if None in self.get_limits():
            self.update_limits(DEFAULT_LIMITS)
        if self.default_value is None:
            self.default_value = DEFAULT_VALUE
            self.update_value(DEFAULT_VALUE)
        self.update_state(self.STATES.READY)

    def get_limits(self):
        my_limits = ActuatorMockup.get_limits(self)
        return my_limits

    def _move(self, value):
        """Simulated energy change
        Args:
            value (float): target energy
        """
        start_pos = self.get_value()
        if value is not None and start_pos is not None:
            step = -1 if value < start_pos else 1
            for _val in range(int(start_pos) + step, int(value) + step, step):
                time.sleep(0.2)
                self.update_value(_val)
        time.sleep(0.2)
        return value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Transmission control for MANACA.
   Attenuator: ABS-300-12-C-DN-DN40 Flanges.
   Foils: #1, 8 um Al;
          #2, 10 um Al;
          #3, 20 um Al;
          #4, 80 um Al;
          #5, 160 um Al;
          #6, 320 um Al;
          #7, 800 um Al;
          #8, 1500 um Al;
          #9, 8 um Ti;
          #10, 10 um Cu
          #11, 5 um Au;
          #12, 25 um Zr.
    Here we use the Beer-Lambert law to calculate the transmission/attenuation: I = I0 * e^-(MU * d)
    were I is the transmitted (final) beam intensity, I0 is the incident (initial) beam intensity,
    MU is the linear attenuation coeficient (cm^-1) and d is the thickness of the foil (cm).
"""
import argparse
import math

from epics import PV


def read_input():
    # Read inputs from terminal
    ap = argparse.ArgumentParser()
    ap.add_argument("-e", "--energy", required=True, type=float, help="Energy of the x-ray beam [keV]")
    args = vars(ap.parse_args())
    return args

def read_transmission(energy):
    # Expressions (interpolation) for calculations of MU (linear attenuation coeficient [cm^-1]) for each material in function of the beam energy.
    # The data used to obtain the expressions (curve fitting) are from  "https://physics.nist.gov/PhysRefData/FFast/html/form.html"
    # This attenuation coeficients can be verified and optimised after experimental validation at the beamline.
    MU_Al = (78657.01011 * math.exp(-energy/0.65969)) + (6406.36151 * math.exp(-energy/1.63268)) + (492.29999 * math.exp(-energy/4.42554)) + (3.2588)
    MU_Ti = (28062.17632 * math.exp(-energy/1.88062)) + (6354120 * math.exp(-energy/0.49973)) + (2257.91488 * math.exp(-energy/5.28296)) + (17.4342)
    MU_Cu = (1147850000 * math.exp(-energy/0.56933)) + (2582.7593 * math.exp(-energy/8.40671)) + (30628.08291 * math.exp(-energy/2.98937)) + (22.4187)
    MU_Au = (42783.04258 * math.exp(-energy/5.39244)) + (399.59714)
    MU_Zr = (11143.28458 * math.exp(-energy/5.87029)) + (102.27474)
    MU_EMPTY = 0.0

    foils_pv = {'F1_Al':(8, MU_Al, PV('MNC:B:RIO01:9425A:bi11'), PV('MNC:B:RIO01:9425A:bi12')),
                'F2_Al':(10, MU_Al, PV('MNC:B:RIO01:9425A:bi10'), PV('MNC:B:RIO01:9425A:bi13')),
                'F3_Al':(20, MU_Al, PV('MNC:B:RIO01:9425A:bi9'), PV('MNC:B:RIO01:9425A:bi14')),
                'F4_Al':(80, MU_Al, PV('MNC:B:RIO01:9425A:bi8'), PV('MNC:B:RIO01:9425A:bi15')),
                'F5_Al':(160, MU_Al, PV('MNC:B:RIO01:9425A:bi7'), PV('MNC:B:RIO01:9425A:bi16')),
                'F6_Al':(320, MU_Al, PV('MNC:B:RIO01:9425A:bi6'), PV('MNC:B:RIO01:9425A:bi17')),
                'F7_Al':(800, MU_Al, PV('MNC:B:RIO01:9425A:bi5'), PV('MNC:B:RIO01:9425A:bi18')),
                'F8_Al':(1500, MU_Al, PV('MNC:B:RIO01:9425A:bi4'), PV('MNC:B:RIO01:9425A:bi19')),
                'F9_Ti':(8, MU_Ti, PV('MNC:B:RIO01:9425A:bi3'), PV('MNC:B:RIO01:9425A:bi20')),
                'F10_Cu':(10, MU_Cu, PV('MNC:B:RIO01:9425A:bi2'), PV('MNC:B:RIO01:9425A:bi21')),
                'F11_Au':(5, MU_Au, PV('MNC:B:RIO01:9425A:bi1'), PV('MNC:B:RIO01:9425A:bi22')),
                'F12_Zr':(25, MU_Zr, PV('MNC:B:RIO01:9425A:bi0'), PV('MNC:B:RIO01:9425A:bi23'))}

    # check what foils are in the the beam and calculate MU*x and then calculate transmission

    # the commented lines bellow works for the attenuator expected logic (0 for foil out and 1 for foil in).
    # as the current logic is inverted (1 is foil out and 0 is foil in) use the uncommented lines.
    foils_in = []
    for foil in foils_pv.keys():
        # if foils_pv[foil][2].get() == 1 and foils_pv[foil][3].get() == 0:
        if foils_pv[foil][2].get() == 0 and foils_pv[foil][3].get() == 1:
            atten_mu_x = float(foils_pv[foil][0] * 1.0E-4) * foils_pv[foil][1]
            foils_in.append(atten_mu_x)
            status = 0
        # elif foils_pv[foil][2].get() == 0 and foils_pv[foil][3].get() == 1:
        elif foils_pv[foil][2].get() == 1 and foils_pv[foil][3].get() == 0:
            status = 0
        else:
            status = 1

    MU_x = sum(foils_in)
    transmission = 1.0 * (math.exp(-MU_x))

    return transmission, status


def main():
    energy = read_input()['energy']
    T = read_transmission(energy)
    transmission = round(T[0] * 100, 2)
    foil_status = T[1]

if __name__ == "__main__":
    main()



import datetime
import json
import logging
import os
import subprocess
import xmlrpc.client

import gevent
from devtools import debug
from pydantic.v1 import (
    BaseModel,
    Field,
)
from typing_extensions import Literal

from mxcubecore import HardwareRepository as HWR
from mxcubecore.queue_entry.base_queue_entry import BaseQueueEntry


class BaseUserCollectionParameters(BaseModel):
    num_images: int = Field(0, description="")
    exp_time: float = Field(100e-6, gt=0, lt=1, description="s")
    sub_sampling: Literal[1, 2, 4, 6, 8] = Field(1)
    take_pedestal: bool = Field(True)

    frequency: float = Field(
        float(HWR.beamline.diffractometer.get_property("max_freq", 925)),
        description="Hz",
    )


class SsxBaseQueueTaskParameters(BaseModel):
    @staticmethod
    def ui_schema():
        return json.dumps(
            {
                "ui:order": [
                    "num_images",
                    "exp_time",
                    "osc_range",
                    "osc_start",
                    "resolution",
                    "transmission",
                    "energy",
                    "vertical_spacing",
                    "horizontal_spacing",
                    "nb_lines",
                    "nb_samples_per_line",
                    "motor_top_left_x",
                    "motor_top_left_y",
                    "motor_top_left_z",
                    "motor_top_right_x",
                    "motor_top_right_y",
                    "motor_top_right_z",
                    "motor_bottom_left_x",
                    "motor_bottom_left_y",
                    "motor_bottom_left_z",
                    "chip_type",
                    "take_pedestal",
                    "sub_sampling",
                    "*",
                ],
                "ui:submitButtonOptions": {
                    "norender": "true",
                },
                "sub_sampling": {"ui:readonly": "true"},
                "frequency": {"ui:readonly": "true"},
            }
        )


class SsxBaseQueueEntry(BaseQueueEntry):
    """
    Defines common SSX collection methods.
    """

    def __init__(self, view, data_model):
        super().__init__(view=view, data_model=data_model)
        self.beamline_values = None

        self._use_nicoproc = False
        self._processing_host = "http://lid29control-2:9998"

    def get_data_path(self):
        data_root_path = HWR.beamline.session.get_image_directory(
            os.path.join(
                self._data_model._task_data.path_parameters.subdir,
                self._data_model._task_data.path_parameters.experiment_name,
            )
        )

        process_path = os.path.join(
            HWR.beamline.session.get_base_process_directory(),
            self._data_model._task_data.path_parameters.subdir,
        )

        return data_root_path, process_path

    def take_pedestal(self, max_freq):
        params = self._data_model._task_data.user_collection_parameters

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        sub_sampling = (
            self._data_model._task_data.user_collection_parameters.sub_sampling
        )

        data_root_path, _ = self.get_data_path()

        packet_fifo_depth = 20000

        if params.take_pedestal:
            HWR.beamline.control.safshut_oh2.close()
            if not hasattr(HWR.beamline.control, "lima2_jungfrau_pedestal_scans"):
                HWR.beamline.control.load_script("id29_lima2.py")

            pedestal_dir = HWR.beamline.detector.find_next_pedestal_dir(
                data_root_path, "pedestal"
            )
            sls_detectors = "/users/blissadm/local/sls_detectors"
            lima2_path = f"{sls_detectors}/lima2"
            cl_source_path = f"{lima2_path}/processings/common/fai/kernels"

            logging.getLogger("user_level_log").info(
                f"Storing pedestal in {pedestal_dir}"
            )
            subprocess.Popen(
                "mkdir --parents %s && chmod -R 755 %s" % (pedestal_dir, pedestal_dir),
                shell=True,
                stdin=None,
                stdout=None,
                stderr=None,
                close_fds=True,
            ).wait()

            HWR.beamline.control.lima2_jungfrau_pedestal_scans(
                HWR.beamline.control.lima2_jungfrau4m_rr_smx,
                exp_time,
                max_freq / sub_sampling,
                1000,
                pedestal_dir,
                "pedestal.h5",
                disable_saving="raw",
                print_params=True,
                det_params={"packet_fifo_depth": packet_fifo_depth},
                cl_source_path=cl_source_path,
            )

            subprocess.Popen(
                "cd %s && rm -f pedestal.h5 && ln -s %s/pedestal.h5"
                % (data_root_path, pedestal_dir),
                shell=True,
                stdin=None,
                stdout=None,
                stderr=None,
                close_fds=True,
            ).wait()

    def start_processing(self, exp_type):
        data_root_path, _ = self.get_data_path()

        if self._use_nicoproc:
            self._start_ssx_processing(
                self.beamline_values,
                self._data_model._task_data,
                data_root_path,
                experiment_type=exp_type,
            )
        else:
            logging.getLogger("user_level_log").info(f"NICOPROC False")

    def prepare_acqiusition(self):
        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        num_images = self._data_model._task_data.user_collection_parameters.num_images
        data_root_path, _ = self.get_data_path()

        HWR.beamline.detector.stop_acquisition()
        HWR.beamline.detector.prepare_acquisition(
            num_images, exp_time, data_root_path, fname_prefix
        )
        HWR.beamline.detector.wait_ready()

    def _monitor_collect(self):
        for i in range(1, 99):
            self.emit_progress(i / 100.0)
            gevent.sleep(0.1)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)

        self._monitor_task = gevent.spawn(self._monitor_collect)

    def pre_execute(self):
        super().pre_execute()
        self.beamline_values = HWR.beamline.lims.pyispyb.get_current_beamline_values()
        self.additional_lims_values = (
            HWR.beamline.lims.pyispyb.get_additional_lims_values()
        )
        self.emit_progress(0)

    def post_execute(self):
        super().post_execute()
        data_root_path, _ = self.get_data_path()
        self.additional_lims_values.end_time = datetime.datetime.now()

        HWR.beamline.lims.pyispyb.create_ssx_collection(
            data_root_path,
            self._data_model._task_data,
            self.beamline_values,
            self.additional_lims_values,
        )

        if HWR.beamline.control.safshut_oh2.state.name == "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.close()

        HWR.beamline.detector.wait_ready()
        HWR.beamline.detector.stop_acquisition()
        self.emit_progress(1)

    def emit_progress(self, progress):
        HWR.beamline.collect.emit_progress(progress)

    def stop(self):
        super().stop()
        if HWR.beamline.control.safshut_oh2.state.name == "OPEN":
            HWR.beamline.control.safshut_oh2.close()
            logging.getLogger("user_level_log").info("shutter closed")

        HWR.beamline.detector.stop_acquisition()

    def _start_processing(self, dc_parameters, file_paramters):
        param = {
            "exposure": dc_parameters["oscillation_sequence"][0]["exposure_time"],
            "detector_distance": dc_parameters["detectorDistance"],
            "wavelength": dc_parameters["wavelength"],
            "orgx": dc_parameters["xBeam"],
            "orgy": dc_parameters["yBeam"],
            "oscillation_range": dc_parameters["oscillation_sequence"][0]["range"],
            "start_angle": dc_parameters["oscillation_sequence"][0]["start"],
            "number_images": dc_parameters["oscillation_sequence"][0][
                "number_of_images"
            ],
            "image_first": dc_parameters["oscillation_sequence"][0][
                "start_image_number"
            ],
            "fileinfo": file_paramters,
        }

        logging.getLogger("HWR").info("NICOPROC START")

        with xmlrpc.client.ServerProxy(self._processing_host) as p:
            p.start(param)

    def _start_ssx_processing(self, beamline_values, params, path, experiment_type=""):
        param = {
            "exposure": params.user_collection_parameters.exp_time,
            "detector_distance": beamline_values.detector_distance,
            "wavelength": beamline_values.wavelength,
            "orgx": beamline_values.beam_x,
            "orgy": beamline_values.beam_y,
            "oscillation_range": params.collection_parameters.osc_range,
            "start_angle": params.collection_parameters.osc_start,
            "number_images": params.user_collection_parameters.num_images,
            "image_first": params.collection_parameters.first_image,
            "fileinfo": params.path_parameters.dict(),
            "root_path": path,
            "experiment_type": experiment_type,
        }

        logging.getLogger("HWR").info("NICOPROC START")

        try:
            with xmlrpc.client.ServerProxy(self._processing_host) as p:
                p.start(param)
        except Exception:
            logging.getLogger("HWR").exception("")

    def _stop_processing(self):
        logging.getLogger("HWR").info("NICOPROC STOP")

        with xmlrpc.client.ServerProxy(self._processing_host) as p:
            p.stop()



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


import logging
import unittest

from mxcubecore.BaseHardwareObjects import HardwareObject

BEAMLINE = None


class TestException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class TestMethods(unittest.TestCase):
    def test_has_channels(self):
        logging.getLogger("HWR").debug("UnitTest: Testing has channels...")

    def test_get_value(self):
        logging.getLogger("HWR").debug("UnitTest: Testing return values...")
        self.assertIn(
            type(BEAMLINE.energy.get_value()),
            (float, int),
            "Energy hwobj | get_current_energy() returns float",
        )

        logging.getLogger("HWR").debug("UnitTest: Testing transmission hwobj")
        self.assertIn(
            type(BEAMLINE.transmission.get_value()),
            (float, int),
            "Transmission hwobj | get_value() returns float",
        )

        logging.getLogger("HWR").debug("UnitTest: Testing aperture hwobj")
        self.assertIn(
            type(BEAMLINE.beam.aperture.get_diameter_size()),
            (float, int),
            "Aperture | get_diameter_size() returns float",
        )
        self.assertIn(
            type(BEAMLINE.beam.aperture.get_diameter_size_list()),
            (list, tuple),
            "Aperture | get_diameter_size_list() returns list or tuple",
        )
        self.assertIn(
            type(BEAMLINE.beam.aperture.get_position_list()),
            (list, tuple),
            "Aperture | get_position_list() returns list or tuple",
        )

    def test_get_limits(self):
        logging.getLogger("HWR").debug("UnitTest: Testing limits...")
        self.assertIsInstance(
            BEAMLINE.energy.get_limits(),
            list,
            "Energy hwobj | get_energy_limits() returns list with two floats",
        )

    def test_get_state(self):
        logging.getLogger("HWR").debug("UnitTest: Testing states...")
        self.assertIsInstance(
            BEAMLINE.transmission.getAttState(),
            str,
            "Transmission hwobj | getAttState() returns int",
        )


class UnitTest(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

    def init(self):
        global BEAMLINE
        from mxcubecore import HardwareRepository as HWR

        BEAMLINE = HWR.beamline

        suite = unittest.TestLoader().loadTestsFromTestCase(TestMethods)
        test_result = unittest.TextTestRunner(verbosity=3).run(suite)

        # test_result.errors
        # test_result.failures
        # test_result.skipped



import logging
import os
import time

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class SOLEILRuche(HardwareObject):
    def __init__(self, *args, **kwargs):
        HardwareObject.__init__(self, *args, **kwargs)

    def init(self):
        self.sync_dir = self.get_property("sync_dir")

    def trigger_sync(self, path):

        try:
            logging.getLogger().info("<SOLEIL Ruche> - trigger_sync path %s." % path)
            logging.getLogger().info(
                "<SOLEIL Ruche> username: %s  user_id: %s projuser: %s"
                % (
                    HWR.beamline.session.username,
                    HWR.beamline.session.user_id,
                    HWR.beamline.session.projuser,
                )
            )
            if HWR.beamline.session.user_id is None:
                return
        except Exception:
            pass
        if os.path.isdir(path):
            path_to_sync = path
        elif os.path.exists(path):
            path_to_sync = os.path.dirname(os.path.abspath(path))
        else:
            logging.getLogger().info(
                "<SOLEIL Ruche> - sync on non existant path %s. Ignored" % path
            )
            path_to_sync = os.path.dirname(os.path.abspath(path))
            # return

        logging.getLogger().info(
            "<SOLEIL Ruche> - triggering data sync on directory %s" % path_to_sync
        )
        ruche_info = HWR.beamline.session.get_ruche_info(path_to_sync)
        try:
            sync_filename = time.strftime(
                "%Y_%m_%d-%H_%M_%S", time.localtime(time.time())
            )
            sync_file_path = os.path.join(self.sync_dir, sync_filename)
            open(sync_file_path, "w").write(ruche_info)
        except Exception:
            logging.getLogger().error(
                "<SOLEIL Ruche> - Cannot write sync in path %s." % sync_file_path
            )


def test():
    import sys

    hwr = HWR.get_hardware_repository()
    hwr.connect()

    ruche = hwr.get_hardware_object("/ruche")
    filename = sys.argv[1]
    ruche.trigger_sync(filename)


if __name__ == "__main__":
    test()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"
__credits__ = ["DESY P11"]
__category__ = "Motor"

import time

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor


class P11DetectorDistance(AbstractMotor):
    def __init__(self, name):
        AbstractMotor.__init__(self, name)

        self.interlock_set = None
        self.limits = None

        self.chan_position = None
        self.chan_state = None
        self.chan_min_value = None
        self.chan_max_value = None
        self.chan_interlock_state = None
        self.cmd_stop = None

    def init(self):
        self.chan_state = self.get_channel_object("axisState")
        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self._set_state)
        self._set_state()

        self.chan_position = self.get_channel_object("axisPosition")
        if self.chan_position is not None:
            self.chan_position.connect_signal("update", self.update_value)
        self.update_value()

        self.chan_min_value = self.get_channel_object("axisMinValue")
        self.chan_max_value = self.get_channel_object("axisMaxValue")

        self.chan_interlock_state = self.get_channel_object("interlockState")
        if self.chan_interlock_state is not None:
            self.chan_interlock_state.connect_signal(
                "update", self.interlock_state_changed
            )
        self.interlock_state_changed()

        self.cmd_stop = self.get_command_object("stopAxis")

        self.chan_actual_value = self.get_channel_object("DistanceLaser")

        # if self.cmd_stop:
        # self.cmd_stop.connect_signal("connected", self.connected)
        # self.cmd_stop.connect_signal("disconnected", self.disconnected)

    def connectNotify(self, signal):
        """
        :param signal: signal
        :type signal: signal
        """
        if signal == "stateChanged":
            self.update_state()
        elif signal == "limitsChanged":
            self.update_limits()
        elif signal == "valueChanged":
            self.update_value()

    def connected(self):
        """
        Sets ready
        :return:
        """
        self._set_state()
        # self.update_state(self.STATES.READY)

    def disconnected(self):
        """
        Sets not ready
        :return:
        """
        self.update_state(self.STATES.OFF)

    def _set_state(self, state=None):

        if state is None:
            _state = self.chan_state.get_value()
        else:
            _state = state

        _state = str(_state)

        if self.interlock_set is None:
            self.update_interlock_state()

        if not self.interlock_set:
            state = self.STATES.FAULT
            self.log.debug(
                "P11 Detector Distance is FAULT because interlock is not set"
            )
        else:
            if _state == "ON":
                state = self.STATES.READY
            elif _state == "MOVING":
                state = self.STATES.BUSY
            else:
                state = self.STATES.FAULT

        self.update_state(state)
        return state

    def get_value(self):
        return self.chan_position.get_value()

    def update_value(self, value=None):
        """Updates motor position
        """
        if value is None:
            value = self.chan_position.get_value()

        super().update_value(value)

    def _set_value(self, value):
        """
        Main move method
        :param value: float
        :return:
        """
        # if self.chan_state is not None:
        # self.update_state(self.STATES.BUSY)

        self.chan_position.set_value(value)

        # Wait until motor is reachiung the actual distance within tolerance.
        tolerance = 1.0  # mm Actual tolerance is within 0.3 range.
        while abs(self.get_value() - value) >= tolerance:
            _state = self.chan_state.get_value()
            if _state == "ON":
                state = self.STATES.READY
                self.update_state(state)
                break
            elif _state == "MOVING":
                state = self.STATES.BUSY
            else:
                state = self.STATES.FAULT
            self.update_state(state)
            time.sleep(0.5)

    def get_limits(self):
        min_value = self.chan_min_value.get_value()
        max_value = self.chan_max_value.get_value()
        return [min_value, max_value]

    def abort(self):
        """Stops motor movement
        """
        self.cmd_stop()

    def get_motor_mnemonic(self):
        """
        Returns motor mnemonic
        :return:
        """
        return "DetectorDistance"

    def update_interlock_state(self, state=None):
        if self.chan_interlock_state is None:
            return

        if state is None:
            state = self.chan_interlock_state.get_value()
        self.interlock_set = state
        self.log.debug(
            "P11 DetectorDistance / INTERLOCK is %s"
            % (self.interlock_set and "SET" or "NOT SET")
        )

    def interlock_state_changed(self, state=None):
        self.update_interlock_state(state)
        self._set_state()



import time

from PyTango.gevent import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task


class ID30A3PhotonFlux(HardwareObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        controller = self.get_object_by_role("controller")
        self.musst = controller.musst
        self.shutter = self.get_deviceby_role("shutter")
        self.factor = self.get_property("current_photons_factor")

        self.shutter.connect("shutterStateChanged", self.shutterStateChanged)

        self.tg_device = DeviceProxy(self.get_property("tango_device"))
        self.counts_reading_task = self._read_counts_task(wait=False)

    @task
    def _read_counts_task(self):
        old_counts = None
        while True:
            counts = self._get_counts()
            if counts != old_counts:
                old_counts = counts
                self.countsUpdated(counts)
            time.sleep(1)

    def _get_counts(self):
        """gain = 20 #20uA
        keithley_voltage = 2
        musst_voltage = int("".join([x for x in self.musst.putget("#?CHCFG CH5") if x.isdigit()]))
        musst_fs = float(0x7FFFFFFF)
        counts = abs((gain/keithley_voltage)*musst_voltage*int(self.musst.putget("#?VAL CH5"))/musst_fs)
        """
        # try:
        self.tg_device.MeasureSingle()
        counts = abs(self.tg_device.ReadData) * 1e6
        return counts

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emitValueChanged()

    def shutterStateChanged(self, _):
        self.countsUpdated(self._get_counts())

    def updateFlux(self, _):
        self.countsUpdated(self._get_counts(), ignore_shutter_state=True)

    def countsUpdated(self, counts, ignore_shutter_state=False):
        if not ignore_shutter_state and self.shutter.getShutterState() != "opened":
            self.emitValueChanged(0)
            return
        flux = counts * self.factor
        self.emitValueChanged("%1.3g" % flux)

        """
        try:
          counts = counts[self.index]
        except TypeError:
          logging.getLogger("HWR").error("%s: counts is None", self.name())
          return
        flux = None

        try:
          egy = HWR.beamline.energy.get_value()*1000.0
        except:
          logging.getLogger("HWR").exception("%s: could not get energy", self.name())
        else:
          try:
            calib_dict = self.calibration_chan.get_value()
            if calib_dict is None:
              logging.getLogger("HWR").error("%s: calibration is None", self.name())
            else:
              calibs = [(float(c["energy"]), float(c[self.counter])) for c in calib_dict.values()]
              calibs.sort()
              E = [c[0] for c in calibs]
              C = [c[1] for c in calibs]
          except:
            logging.getLogger("HWR").exception("%s: could not get calibration", self.name())
          else:
            try:
              aperture_coef = self.aperture.getApertureCoef()
            except:
              aperture_coef = 1
            if aperture_coef <= 0:
              aperture_coef = 1
            calib = numpy.interp([egy], E, C)
            flux = counts * calib * aperture_coef
            #logging.getLogger("HWR").debug("%s: flux-> %f * %f=%f , calib_dict=%r", self.name(), counts, calib, counts*calib, calib_dict)
            self.emitValueChanged("%1.3g" % flux)
        """

    def get_value(self):
        return self.current_flux

    def emitValueChanged(self, flux=None):
        if flux is None:
            self.current_flux = None
            self.emit("valueChanged", ("?",))
        else:
            self.current_flux = flux
            self.emit("valueChanged", (self.current_flux,))



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

try:
    from SpecClient_gevent import SpecVariable
    from SpecClient_gevent.SpecCommand import SpecCommandA
    from SpecClient_gevent.SpecVariable import SpecVariableA
except ImportError:
    from SpecClient.SpecCommand import SpecCommandA
    from SpecClient.SpecVariable import SpecVariableA
    from SpecClient import SpecVariable

from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)

__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


class SpecCommand(CommandObject, SpecCommandA):
    def __init__(self, name, command, version=None, username=None, **kwargs):
        CommandObject.__init__(self, name, username, **kwargs)
        SpecCommandA.__init__(self, command, version)
        self.__cmdExecution = False

    def setSpecVersion(self, version):
        self.connectToSpec(version)

    def replyArrived(self, reply):
        SpecCommandA.replyArrived(self, reply)

        self.__cmdExecution = False

        if reply.error:
            self.emit("commandFailed", (reply.error_code, str(self.name())))
        else:
            self.emit("commandReplyArrived", (reply.get_value(), str(self.name())))

    def beginWait(self):
        self.__cmdExecution = True
        self.emit("commandBeginWaitReply", (str(self.name()),))

    def abort(self):
        SpecCommandA.abort(self)

        self.__cmdExecution = False
        self.emit("commandAborted", (str(self.name()),))

    def is_connected(self):
        return SpecCommandA.isSpecConnected(self)

    def connected(self):
        self.__cmdExecution = False
        self.emit("connected", ())

    def disconnected(self):
        if self.__cmdExecution:
            self.__cmdExecution = False
            self.emit("commandFailed", (-1, str(self.name())))

        self.emit("disconnected", ())
        self.statusChanged(ready=False)

    def statusChanged(self, ready):
        if ready:
            self.emit("commandReady", ())
        else:
            self.emit("commandNotReady", ())


class SpecChannel(ChannelObject, SpecVariableA):
    def __init__(
        self,
        name,
        varname,
        version=None,
        username=None,
        dispatchMode=SpecVariable.FIREEVENT,
        **kwargs
    ):
        ChannelObject.__init__(self, name, username, **kwargs)
        SpecVariableA.__init__(self, varname, version, dispatchMode)

    def setSpecVersion(self, version):
        self.connectToSpec(version)

    def update(self, value):
        ChannelObject.update(self, value)
        self.emit("update", (value,))

    def connected(self):
        self.emit("connected", ())

    def disconnected(self):
        self.emit("disconnected", ())

    def is_connected(self):
        return SpecVariableA.isSpecConnected(self)

    def get_value(self):
        return SpecVariableA.get_value(self)



"""Tango Shutter Hardware Object
Example XML::

  <object class="ALBAEpsActuator">
    <username>Photon Shutter</username>
    <taurusname>bl13/ct/eps-plc-01</taurusname>
    <channel type="sardana" polling="events" name="actuator">pshu</channel>
    <states>Open,Closed</states>
  </object>


Public Interface:
   Commands:
       int get_state()
           Description:
               returns current state
           Output:
               integer value describing the state
               current states correspond to:
                      0: out
                      1: in
                      9: moving
                     11: alarm
                     13: unknown
                     23: fault

       string getStatus()
           Description:
               returns current state as a string that can contain a more
               descriptive information about current state

           Output:
               status string

       cmdIn()
           Executes the command associated to the "In" action
       cmdOut()
           Executes the command associated to the "Out" action

   Signals:
       stateChanged

"""

import logging

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR

STATE_OUT, STATE_IN, STATE_MOVING, STATE_FAULT, STATE_ALARM, STATE_UNKNOWN = (
    0,
    1,
    9,
    11,
    13,
    23,
)


class ALBAEpsActuator(BaseHardwareObjects.HardwareObject):

    states = {
        STATE_OUT: "out",
        STATE_IN: "in",
        STATE_MOVING: "moving",
        STATE_FAULT: "fault",
        STATE_ALARM: "alarm",
        STATE_UNKNOWN: "unknown",
    }

    default_state_strings = ["Out", "In"]

    def __init__(self, name):
        super().__init__(name)

    def init(self):
        self.actuator_state = STATE_UNKNOWN

        try:
            self.actuator_channel = self.get_channel_object("actuator")
            self.actuator_channel.connect_signal("update", self.stateChanged)
        except KeyError:
            logging.getLogger().warning(
                "%s: cannot report EPS Actuator State", self.name()
            )

        try:
            state_string = self.get_property("states")
            if state_string is None:
                self.state_strings = self.default_state_strings
            else:
                states = state_string.split(",")
                self.state_strings = states[1].strip(), states[0].strip()
        except Exception:
            import traceback

            logging.getLogger("HWR").warning(traceback.format_exc())
            self.state_strings = self.default_state_strings

    def get_state(self):
        state = self.actuator_channel.get_value()
        self.actuator_state = self.convert_state(state)
        return self.actuator_state

    def convert_state(self, state):
        if state == 0:
            act_state = STATE_OUT
        elif state == 1:
            act_state = STATE_IN
        else:
            act_state = STATE_UNKNOWN
        return act_state

    def stateChanged(self, value):
        #
        # emit signal
        #
        self.actuator_state = self.convert_state(value)
        self.emit("stateChanged", ((self.actuator_state),))

    def getUserName(self):
        return self.username

    def getStatus(self):
        """
        """
        state = self.get_state()

        if state in [STATE_OUT, STATE_IN]:
            return self.state_strings[state]
        elif state in self.states:
            return self.states[state]
        else:
            return "Unknown"

    def open(self):
        self.cmdIn()

    def close(self):
        self.cmdOut()

    def cmdIn(self):
        self.actuator_channel.set_value(1)

    def cmdOut(self):
        self.actuator_channel.set_value(0)


def test_hwo(hwo):
    print("Name is: ", hwo.getUserName())
    print("Shutter state is: ", hwo.get_state())
    print("Shutter status is: ", hwo.getStatus())

    # print "Opening it"
    # print hwo.open()
    # print "Closing it"
    # print hwo.close()



from . import ID30BeamCmds


class ID30A1BeamCmds(ID30BeamCmds.ID30BeamCmds):
    def __init__(self, *args):
        ID30BeamCmds.ID30BeamCmds.__init__(self, *args)

    def init(self):
        controller = self.get_object_by_role("controller")
        self.centrebeam = ID30BeamCmds.ControllerCommand(
            "Centre beam", controller.centrebeam
        )

    def get_commands(self):
        return [self.centrebeam]



# -*- coding: utf-8 -*-

import inspect
import logging

import jsonpickle

from mxcubecore import HardwareRepository as HWR
from mxcubecore.model import (
    queue_model_enumerables,
    queue_model_objects,
)

xmlrpc_prefix = ""


def queue_set_serialisation(self, backend):
    if backend.lower() == "json":
        return True
    else:
        raise ValueError("Unknown backend type '%s'" % backend)


def queue_get_serialisation(self):
    return "json"


def queue_get_available_serialisations(self):
    """
    Returns a tuple of the available serialisation methods for
    native queue objects
    """
    return ("json",)


def queue_add_node(server_hwobj, task_node, set_on=True):
    """
    Adds the TaskNode objects contained in the json seralized
    list of TaskNodes passed in <task_node>.

    The TaskNodes are marked as activated in the queue if <set_on>
    is True and to inactivated if False.

    :param task_node: TaskNode object to add to queue
    :type parent: TaskNode

    :param set_on: Mark TaskNode as activated if True and as inactivated
                   if false.
    :type set_on: bool

    :returns: True on success otherwise False
    :rtype: bool
    """

    try:
        task = jsonpickle.decode(task_node)
    except Exception as ex:
        logging.getLogger("HWR").exception(str(ex))
        raise

    server_hwobj._add_to_queue(task, set_on)
    return True


def queue_add_child(server_hwobj, parent_id, child):
    """
    Adds the model node task to parent_id.

    :param parent_id: The id of the parent.
    :type parent_id: int

    :param child: The TaskNode object to add.
    :type child: TaskNode

    :returns: The id of the added TaskNode object.
    :rtype: int
    """

    try:
        task = jsonpickle.decode(child)
    except Exception as ex:
        logging.getLogger("HWR").exception(str(ex))
        raise

    node_id = server_hwobj._model_add_child(parent_id, task)
    return node_id


def queue_get_node(server_hwobj, node_id):
    """
    :returns the TaskNode object with the node id <node_id>
    :rtype: TaskNode
    """
    node = server_hwobj._model_get_node(node_id)
    return jsonpickle.encode(node)


def queue_update_result(server_hwobj, node_id, html_report):
    result = False
    node = server_hwobj._model_get_node(node_id)

    if isinstance(node, queue_model_objects.DataCollection):
        node.html_report = str(html_report)
        result = True

    return result


def queue_get_full_path(server_hwobj, subdir, tag):
    """ """
    return HWR.beamline.session.get_full_path(subdir, tag)


def queue_get_model_code(server_hwobj):
    """
    returns a list of tuples of (name of queue model module, source code of queue model).

    The client can compile and use the queue model as follows:

        for (module_name, module_code) in server.native_get_queue_model_code():
            queue_model_objects = imp.new_module(module_name)
            exec module_code in queue_model_objects.__dict__
            sys.modules[module_name] = queue_model_objects

    The module containing the required queue model objects is the last one in
    the list, so after the loop exits queue_model_objects is set to the module
    that is needed by the client.

    Modules other than the last one are modules imported by the
    queue model that are not available for direct import by the XML-RPC client
    """

    # Recipe from
    # http://code.activestate.com/recipes/82234-importing-a-dynamically-generated-module/

    # At the moment, queue_model_objects does not import anything except
    # standard Python modules, so we only need to send over the code for the
    # queue model itself

    queue_model_enumerables_code = inspect.getsource(queue_model_enumerables)

    queue_model_objects_code = inspect.getsource(queue_model_objects)
    queue_model_objects_code = queue_model_objects_code.replace(
        "from mxcubecore import HardwareRepository as HWR",
        "HWR = None  # HardwareRepository not available in this context",
    )

    return [
        (queue_model_enumerables.__name__, queue_model_enumerables_code),
        (queue_model_objects.__name__, queue_model_objects_code),
    ]



"""
A client for PyISPyB Webservices.
"""

import datetime
import json
import logging
import os

import pyispyb_client
from pyispyb_client import Configuration
from pyispyb_client.apis.tags import (
    authentication_api,
    serial_crystallography_api,
    webservices_serial_crystallography_api,
)
from pyispyb_client.model.event_chain_create import EventChainCreate
from pyispyb_client.model.event_create import EventCreate
from pyispyb_client.model.login import Login
from pyispyb_client.model.ssx_crystal_create import SSXCrystalCreate
from pyispyb_client.model.ssx_data_collection_create import SSXDataCollectionCreate
from pyispyb_client.model.ssx_data_collection_group_create import (
    SSXDataCollectionGroupCreate,
)
from pyispyb_client.model.ssx_protein_create import SSXProteinCreate
from pyispyb_client.model.ssx_sample_component_create import SSXSampleComponentCreate
from pyispyb_client.model.ssx_sample_create import SSXSampleCreate
from typing_extensions import Literal

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model.common import (
    BeamlineParameters,
    ISPYBCollectionPrameters,
)


class PyISPyBClient(HardwareObject):
    """
    PyISPyB Web-service client.
    """

    def __init__(self, name):
        HardwareObject.__init__(self, name)

    def init(self):
        self._token_timestamp = None

        if HWR.beamline.session:
            self.beamline_name = HWR.beamline.session.beamline_name
        else:
            self.beamline_name = "ID:TEST"

        self._username = self.get_property("username", "").strip()
        self._password = self.get_property("password", "").strip()
        self._plugin = self.get_property("plugin", "").strip()
        self._host = self.get_property("host", "").strip()

        self._configuration = Configuration(
            host=self._host,
        )

        self._update_token()

    def _update_token(self, request_token=False):
        """ """
        if not self._token_timestamp:
            request_token = True
        else:
            timedelta = datetime.datetime.now() - self._token_timestamp
            if timedelta.seconds > (3 * 60 * 60):
                request_token = True

        if request_token:
            try:
                self.authenticate()
            except Exception:
                logging.getLogger("HWR").exception("")

    def authenticate(self):
        with pyispyb_client.ApiClient(self._configuration) as api_client:
            api_instance = authentication_api.AuthenticationApi(api_client)
            login = Login(
                plugin=self._plugin, login=self._username, password=self._password
            )

            try:
                api_response = api_instance.login_ispyb_api_v1_auth_login_post(login)

                self._configuration = Configuration(
                    host=self._host,
                    #                   access_token=api_response.body.token,
                )
                self._configuration.access_token = api_response.body.token
            except pyispyb_client.ApiException as e:
                print(
                    "Exception when calling AuthenticationApi->login_ispyb_api_v1_auth_login_post: %s\n"
                    % e
                )

    def get_current_beamline_values(self):
        return BeamlineParameters(
            **{
                "energy": HWR.beamline.energy.get_value(),
                "wavelength": HWR.beamline.energy.get_wavelength(),
                "resolution": HWR.beamline.resolution.get_value(),
                "transmission": HWR.beamline.transmission.get_value(),
                "detector_distance": HWR.beamline.detector.distance.get_value(),
                "beam_x": HWR.beamline.detector.get_beam_position()[0],
                "beam_y": HWR.beamline.detector.get_beam_position()[1],
                "beam_size_x": HWR.beamline.beam.get_beam_size()[0],
                "beam_size_y": HWR.beamline.beam.get_beam_size()[1],
                "beam_shape": HWR.beamline.beam.get_beam_shape(),
                "energy_bandwidth": HWR.beamline.beam.get_property(
                    "energy_bandwidth", 0.1
                ),
            }
        )

    def get_additional_lims_values(self):
        return ISPYBCollectionPrameters(
            **{
                "flux_start": HWR.beamline.flux.get_value(),
                "flux_end": HWR.beamline.flux.get_value(),
                "start_time": datetime.datetime.now(),
                "end_time": datetime.datetime.now(),
                "chip_model": HWR.beamline.collect.get_property("chip_model", ""),
                "polarisation": HWR.beamline.beam.polarisation,
                "mono_stripe": HWR.beamline.beam.get_property("mono_stripe", ""),
            }
        )

    def mxcube_to_ispyb_collection_type(
        self, collection_type: str
    ) -> Literal["SSX-CHIP", "SSX-Jet"]:
        val = "SSX-CHIP"

        if collection_type == "ssx_chip_collection":
            val = "SSX-CHIP"
        elif collection_type == "ssx_injector_collection":
            val = "SSX-Jet"

        return val

    def create_ssx_data_collection_group(self, collection_parameters, session_id=None):
        self._update_token()
        sacronym, sname = collection_parameters.path_parameters.prefix.split("-")
        session_id = session_id if session_id else int(HWR.beamline.session.session_id)

        with pyispyb_client.ApiClient(self._configuration) as api_client:
            # Create an instance of the API class
            api_instance = webservices_serial_crystallography_api.WebservicesSerialCrystallographyApi(
                api_client
            )
            ssx_data_collection_group_create = {
                "sessionId": session_id,
                "startTime": datetime.datetime.now(),
                # "endTime": datetime.datetime.now(),
                "experimentType": self.mxcube_to_ispyb_collection_type(
                    collection_parameters.common_parameters.type
                ),
                "experimentName": collection_parameters.common_parameters.label,
                #                "comments": "comments_example",
                "sample": {
                    "name": sname,
                    "support": "support_example",
                    "crystal": {
                        #                        "size_X": -1.0,
                        #                        "size_Y": -1.0,
                        #                        "size_Z": -1.0,
                        #                        "abundance": -1.0,
                        "protein": {
                            "name": sname,
                            "acronym": sacronym,
                        },
                        "components": [],
                        #                      "components": [{
                        #                               "name": "name",
                        #                               "componentType": "Ligand",
                        #                               "composition": "composition_example",
                        #    "abundance": -1.0,
                        #                           },
                        #                       ],
                        #                   },"components":[{
                        #                           "name": "name",
                        #                           "componentType": "Ligand",
                        #                           "composition": "composition_example",
                        #                            "abundance": -1.0,
                        #                       },
                        #                    ],
                    },
                    "components": [],
                },
            }

            try:
                api_response = api_instance.create_datacollectiongroup(
                    ssx_data_collection_group_create
                )
                # pprint.pprint(api_response)
                return int(api_response.body)
            except pyispyb_client.ApiException as e:
                print(
                    "Exception when calling SerialCrystallographyApi->create_datacollectiongroup: %s\n"
                    % e
                )

    def create_ssx_data_collection(
        self, dcg_id, collection_parameters, beamline_parameters, extra_lims_values
    ):
        # Enter a context with an instance of the API client
        with pyispyb_client.ApiClient(self._configuration) as api_client:
            # Create an instance of the API class
            api_instance = webservices_serial_crystallography_api.WebservicesSerialCrystallographyApi(
                api_client
            )
            ssx_data_collection_create = {
                "dataCollectionGroupId": dcg_id,
                "exposureTime": collection_parameters.user_collection_parameters.exp_time,
                "transmission": beamline_parameters.transmission,
                "flux": extra_lims_values.flux_start,
                "xBeam": beamline_parameters.beam_x,
                "yBeam": beamline_parameters.beam_y,
                "wavelength": beamline_parameters.wavelength,
                "detectorDistance": beamline_parameters.detector_distance,
                "beamSizeAtSampleX": beamline_parameters.beam_size_x,
                "beamSizeAtSampleY": beamline_parameters.beam_size_y,
                "average_temperature": 0.0,
                "xtalSnapshotFullPath1": "",
                "xtalSnapshotFullPath2": "",
                "xtalSnapshotFullPath3": "",
                "xtalSnapshotFullPath4": "",
                "imagePrefix": collection_parameters.path_parameters.prefix,
                "numberOfPasses": 1,
                "numberOfImages": collection_parameters.collection_parameters.num_images,
                "resolution": beamline_parameters.resolution,
                "resolutionAtCorner": 0.0,
                "flux_end": extra_lims_values.flux_end,
                "detector_id": HWR.beamline.detector.get_property("detector_id"),
                "startTime": extra_lims_values.start_time,
                "endTime": extra_lims_values.end_time,
                "repetitionRate": 0.0,
                "energyBandwidth": beamline_parameters.energy_bandwidth,
                "monoStripe": extra_lims_values.mono_stripe,
                "jetSize": 0,
                "jetSpeed": 0,
                "laserEnergy": 0,
                "chipModel": extra_lims_values.chip_model,
                "chipPattern": "",
                "beamShape": beamline_parameters.beam_shape,
                "polarisation": extra_lims_values.polarisation,
                "underlator_gap1": 0,
                "event_chains": [
                    {
                        "name": "name",
                        "events": [
                            {
                                "type": "XrayDetection",
                                "name": "name",
                                "offset": 0.0,
                                "duration": 0.0,
                                "period": 0.0,
                                "repetition": 0.0,
                            },
                        ],
                    },
                ],
            }

            # example passing only required values which don't have defaults set
            try:
                # Create Datacollection
                api_response = api_instance.create_datacollection(
                    ssx_data_collection_create
                )
                import pprint

                pprint.pprint(api_response)
            except pyispyb_client.ApiException as e:
                print(
                    "Exception when calling SerialCrystallographyApi->create_datacollection: %s\n"
                    % e
                )

    def create_ssx_collection(
        self, data_path, collection_parameters, beamline_parameters, extra_lims_values
    ):
        try:
            fpath = os.path.abspath(os.path.join(data_path + "../", "ispyb_data.json"))

            if os.path.exists(fpath):
                with open(fpath, "r+") as _f:
                    data = json.loads(fpath.read())
                    dcg_id = data["dcg_id"]
            else:
                dcg_id = self.create_ssx_data_collection_group(collection_parameters)

                with open(fpath, "w+") as _f:
                    data = {"dcg_id": dcg_id}
                    str_data = json.dumps(data, indent=4)
                    _f.write(str_data)

                self.create_ssx_data_collection(
                    dcg_id,
                    collection_parameters,
                    beamline_parameters,
                    extra_lims_values,
                )
        except Exception:
            logging.getLogger("HWR").exception("")



import logging
import os
import subprocess
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.CommandContainer import ConnectionError
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector


class LimaJungfrauDetector(AbstractDetector):
    def __init__(self, name):
        AbstractDetector.__init__(self, name)
        self.header = dict()
        self.start_angles = list()

    def init(self):
        AbstractDetector.init(self)
        lima_device = self.get_property("lima_device", "")

        if not lima_device:
            return

        t = lima_device.split("/")
        t[-2] = "mask"
        mask_device = "/".join(t)

        try:
            for channel_name in (
                "latency_time",
                "State",
                "acq_status",
                "acq_trigger_mode",
                "saving_mode",
                "acq_nb_frames",
                "acq_expo_time",
                "saving_directory",
                "saving_prefix",
                "saving_suffix",
                "saving_next_number",
                "saving_index_format",
                "saving_format",
                "saving_overwrite_policy",
                "saving_header_delimiter",
                "saving_frame_per_file",
                "last_image_saved",
                "image_roi",
            ):
                self.add_channel(
                    {"type": "tango", "name": channel_name, "tangoname": lima_device},
                    channel_name,
                )

            self.add_channel(
                {"type": "tango", "name": "mask_file", "tangoname": mask_device},
                "MaskFile",
            )
            self.add_channel(
                {"type": "tango", "name": "mask_run_level", "tangoname": mask_device},
                "RunLevel",
            )
            self.add_command(
                {"type": "tango", "name": "start_mask", "tangoname": mask_device},
                "Start",
            )
            self.add_command(
                {"type": "tango", "name": "stop_mask", "tangoname": mask_device},
                "Stop",
            )

            self.add_command(
                {"type": "tango", "name": "prepare_acq", "tangoname": lima_device},
                "prepareAcq",
            )
            self.add_command(
                {"type": "tango", "name": "start_acq", "tangoname": lima_device},
                "startAcq",
            )
            self.add_command(
                {"type": "tango", "name": "stop_acq", "tangoname": lima_device},
                "stopAcq",
            )
            self.add_command(
                {"type": "tango", "name": "reset", "tangoname": lima_device}, "reset"
            )
            self.add_command(
                {"type": "tango", "name": "set_image_header", "tangoname": lima_device},
                "SetImageHeader",
            )

            # prepareAcq can block after large data transfers (buffered by GPFS): 2 min timeout
            self.get_command_object("prepare_acq").set_device_timeout(2 * 60 * 1000)
            self._emit_status()

        except ConnectionError:
            self.update_state(HardwareObjectState.FAULT)
            logging.getLogger("HWR").error(
                "Could not connect to detector %s" % lima_device
            )
            self._emit_status()

    def has_shutterless(self):
        return True

    def wait_ready(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not ready")):
            while self.get_channel_value("acq_status") != "Ready":
                time.sleep(1)

    def last_image_saved(self):
        try:
            img = self.get_channel_object("last_image_saved").get_value() + 1
            return img
        except Exception:
            return 0

    def get_deadtime(self):
        return float(self.get_property("deadtime"))

    def prepare_acquisition(self, number_of_images, exptime, data_root_path, prefix):
        self.set_channel_value("acq_trigger_mode", "EXTERNAL_TRIGGER_MULTI")

        self.set_channel_value("acq_nb_frames", number_of_images)
        self.set_channel_value("acq_expo_time", exptime)
        self.set_channel_value("latency_time", 990e-6)
        self.set_channel_value("saving_frame_per_file", 1000)

        mask_file = self.get_property("mask_file", None)

        if mask_file:
            self.set_channel_value("mask_file", mask_file)

        self.set_channel_value("mask_run_level", 0)

        self.set_detector_filenames(data_root_path, prefix)

    def find_next_pedestal_dir(self, data_root_path, subdir):
        _index = 1
        _indes_str = "%04d" % _index
        fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        while os.path.exists(fpath):
            _index += 1
            _indes_str = "%04d" % _index
            fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        return fpath

    def set_detector_filenames(self, data_root_path, prefix):
        subprocess.Popen(
            "mkdir --parents %s" % (data_root_path),
            shell=True,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        ).wait()

        subprocess.Popen(
            "chmod -R 755 %s" % (data_root_path),
            shell=True,
            stdin=None,
            stdout=None,
            stderr=None,
            close_fds=True,
        ).wait()

        self.set_channel_value("saving_mode", "AUTO_FRAME")
        self.set_channel_value("saving_directory", data_root_path)
        self.set_channel_value("saving_prefix", prefix)
        self.set_channel_value("saving_format", "HDF5BS")

    def start_acquisition(self):
        self.wait_ready()
        self.execute_command("stop_acq")
        self.execute_command("start_mask")
        self.execute_command("prepare_acq")
        self.execute_command("start_acq")
        self._emit_status()

    def stop_acquisition(self):
        try:
            self.execute_command("stop_acq")
        except Exception:
            pass
        finally:
            self.wait_ready()
            self.execute_command("reset")
            self.wait_ready()
            self.execute_command("stop_mask")

        self._emit_status()

    def reset(self):
        self.stop_acquisition()

    @property
    def status(self):
        try:
            acq_status = self.get_channel_value("acq_status")
        except Exception:
            acq_status = "OFFLINE"

        status = {
            "acq_satus": acq_status.upper(),
        }

        return status

    def _emit_status(self):
        self.emit("statusChanged", self.status)

    def restart(self) -> None:
        self.reset()
        return None



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Mockup file to define _execute_spectrum method
"""

from time import sleep

from mxcubecore.HardwareObjects.abstract.AbstractXRFSpectrum import AbstractXRFSpectrum

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class XRFSpectrumMockup(AbstractXRFSpectrum):
    """Overload the abstract method"""

    def _execute_spectrum(self, integration_time=None, filename=None):
        """Specific XRF acquisition procedure.

        Args:
            integration_time (float): MCA integration time [s].
            filename (str): Data file (full path).
        """
        integration_time = integration_time or self.default_integration_time
        sleep(integration_time)
        return True



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""

--------------------------------------------
Description
--------------------------------------------

AbstractSampleChanger is a base class to help in the implementation of
Hardware Objects for SampleChangers following the
"SampleChanger Standard Interface".

If this class is used as base class a standard class is then provided for
its use by generic bricks or by MXCuBE itself.  This class exposes the
following API for bricks and MXCuBE:

--------------------------------------------
SampleChanger - Standard Interface
--------------------------------------------

Sample Changer States
----------------------

    SampleChangerState.Unknown
    SampleChangerState.Ready
    SampleChangerState.Loaded
    SampleChangerState.Loading
    SampleChangerState.Unloading
    SampleChangerState.Selecting
    SampleChangerState.Scanning
    SampleChangerState.Resetting
    SampleChangerState.Charging
    SampleChangerState.Moving
    SampleChangerState.ChangingMode
    SampleChangerState.StandBy
    SampleChangerState.Disabled
    SampleChangerState.Alarm
    SampleChangerState.Fault
    SampleChangerState.Initializing
    SampleChangerState.Closing

Commands
----------------------

load()
unload()
select()
abort()
change_mode()

get_state()
get_status()
is_ready()
wait_ready()
has_loaded_sample()
get_loaded_sample()

Specifying sample locations
-----------------------------
The sample model in a sample changer is based
in the model:
   SampleChanger
      Container
         [Container...]
            Sample

Typically for a sample changer with Pucks and Sample
there is a single level for Container. Specifying
a sample location will consist in giving the puck (basket)
number followed by the sample number. In the location
example `3:5` the fifth sample in the third puck is specified.

For other more complex constructions (for example for
a plate manipulator) each nested container will be specified
until getting to the sample:

In the example for a location in a plate manipulator like `1:5:2`
the location specifies first plate well, fifth drop, second crystal.


Events emitted
----------------------

SampleChanger.STATE_CHANGED_EVENT
SampleChanger.STATUS_CHANGED_EVENT
SampleChanger.INFO_CHANGED_EVENT
SampleChanger.LOADED_SAMPLE_CHANGED_EVENT
SampleChanger.SELECTION_CHANGED_EVENT
SampleChanger.TASK_FINISHED_EVENT
SampleChanger.PROGRESS_MESSAGE

Tools for SC Classes
----------------------

- useUpdateTimer (xml property):
   This property can accept a boolean value (True/False)

   If this property is set the HardwareObject will
   poll itself for state changes, information change and
   other needed values.

   Include a line like `<useUpdateTimer>True</useUpdateTimer`
   in the xml file



--------------------------------------------
How to implement derived SC Classes
--------------------------------------------

"""

import abc
import logging

from gevent import (
    Timeout,
    sleep,
)

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.sample_changer.Container import Container
from mxcubecore.TaskUtils import task as dtask


class SampleChangerState:
    """
    Enumeration of sample changer states
    """

    Unknown = 0
    Ready = 1
    Loaded = 2
    Loading = 3
    Unloading = 4
    Selecting = 5
    Scanning = 6
    Resetting = 7
    Charging = 8
    Moving = 9
    ChangingMode = 10
    StandBy = 11
    Disabled = 12
    Alarm = 13
    Fault = 14
    Initializing = 15
    Closing = 16

    STATE_DESC = {
        Ready: "Ready",
        Loaded: "Loaded",
        Alarm: "Alarm",
        Charging: "Charging",
        Disabled: "Disabled",
        Fault: "Fault",
        Loading: "Loading",
        Resetting: "Resetting",
        Scanning: "Scanning",
        Selecting: "Selecting",
        Unloading: "Unloading",
        Moving: "Moving",
        ChangingMode: "Changing Mode",
        StandBy: "StandBy",
        Initializing: "Initializing",
        Closing: "Closing",
    }

    @staticmethod
    def tostring(state):
        """Convert state to string"""
        return SampleChangerState.STATE_DESC.get(state, "Unknown")


class SampleChangerMode:
    """
    Enumeration of sample changer operating modes
    """

    Unknown = 0
    Normal = 1
    Charging = 8
    Disabled = 11


class SampleChanger(Container, HardwareObject):
    """
    Abstract base class for sample changers
    """

    __metaclass__ = abc.ABCMeta

    # ########################    EVENTS    #########################
    STATE_CHANGED_EVENT = "stateChanged"
    STATUS_CHANGED_EVENT = "statusChanged"
    INFO_CHANGED_EVENT = "infoChanged"
    LOADED_SAMPLE_CHANGED_EVENT = "loadedSampleChanged"
    SELECTION_CHANGED_EVENT = "selectionChanged"
    TASK_FINISHED_EVENT = "taskFinished"
    CONTENTS_UPDATED_EVENT = "contentsUpdated"
    PROGRESS_MESSAGE = "progress_message"

    def __init__(self, type_, scannable, *args, **kwargs):
        super().__init__(type_, None, type_, scannable)
        if len(args) == 0:
            args = (type_,)
        HardwareObject.__init__(self, *args, **kwargs)
        self.state = -1
        self.status = ""
        self._progress_message = ""
        self._set_state(SampleChangerState.Unknown)
        self.task = None
        self.task_proc = None
        self.task_error = None
        self._transient = False
        self._token = None
        self._timer_update_inverval = 5  # interval in periods of 100 ms
        self._timer_update_counter = 0
        self.use_update_timer = None

    def init(self):
        """
        HardwareObject init method
        """
        use_update_timer = self.get_property("useUpdateTimer", True)

        msg = f"SampleChanger: Using update timer is {use_update_timer}"
        logging.getLogger("HWR").info(msg)

        if use_update_timer:
            task1s = self.__timer_1s_task(wait=False)
            task1s.link(self._on_timer_1s_exit)
            updateTask = self.__update_timer_task(wait=False)
            updateTask.link(self._on_timer_update_exit)

        self.use_update_timer = use_update_timer

        self.update_info()

    def _on_timer_1s_exit(self, task):
        logging.warning("Exiting Sample Changer 1s timer task")

    def _on_timer_update_exit(self, task):
        logging.warning("Exiting Sample Changer update timer task")

    @dtask
    def __timer_1s_task(self, *args):
        while True:
            sleep(1.0)
            try:
                if self.is_enabled():
                    self._on_timer_1s()
            except Exception:
                pass

    @dtask
    def __update_timer_task(self, *args):
        while True:
            sleep(1)
            try:
                if self.is_enabled():
                    self._timer_update_counter += 1
                    if self._timer_update_counter >= self._timer_update_counter:
                        self._on_timer_update()
                        self._timer_update_counter = 0
            except Exception:
                pass

    # ########################    TIMER    #########################
    def _set_timer_update_interval(self, value):
        self._timer_update_inverval = value

    def _on_timer_update(self):
        # if not self.is_executing_task():
        self.update_info()

    def _on_timer_1s(self):
        pass

    # #######################    HardwareObject    #######################

    def connect_notify(self, signal):
        msg = f"connect_notify {signal}"
        logging.getLogger().info(msg)

    # ########################    PUBLIC    #########################

    def get_state(self):
        """
        Returns:
            (SampleChangerState): Current sample changer state
        """
        return self.state

    def get_status(self):
        """
        Returns:
            (str) String representation of current state
        :rtype: str
        """
        return self.status

    @property
    def progress_message(self) -> str:
        """
        Returns:
            Current progress message
        """
        self._progress_message

    def set_progress_message(self, message: str) -> None:
        """
        Set progress message describing the current sample changer activity.

        Args:
            message: string describing current sample changer activity.
        """
        self._progress_message = message
        self._trigger_progress_message(message)

    def get_task_error(self):
        """
        Returns:
            (str): Description of the error of last executed task (or None if success)
        """
        return self.task_error

    def is_ready(self):
        """
        Returns:
            (bool): True if the state corresponds to READY.
        """
        return self.state in (
            SampleChangerState.Ready,
            SampleChangerState.Loaded,
            SampleChangerState.Charging,
            SampleChangerState.StandBy,
        )

    def wait_ready(self, timeout=None):
        """Wait for current sample changer operation to finish.
        Blocks for timeout seconds or forever if timout is None
        Args:
            timeout (int): timeout [s].
        Raises:
            (Exception): If operation lasts longer than the timeout.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting ready")):
            while not self.is_ready():
                sleep(0.5)

    def is_normal_state(self):
        """
        Returns:
            (bool): True if state is not in the 'NOT USABLE' states list.
        """
        return self.state not in (
            SampleChangerState.Disabled,
            SampleChangerState.Alarm,
            SampleChangerState.Fault,
            SampleChangerState.Unknown,
        )

    def is_enabled(self):
        """
        Returns:
            (boolean): True if sample changer is enabled otherwise False
        """
        return self.state != SampleChangerState.Disabled

    def assert_enabled(self):
        """
        Raises:
            (Exception): If sample changer is disabled.
        """
        if not self.is_enabled():
            raise Exception("Sample Changer is disabled")

    def assert_not_charging(self):
        """
        Raises:
            (Exception): If sample changer is not charging
        """
        if self.state == SampleChangerState.Charging:
            raise Exception("Sample Changer is in Charging mode")

    def assert_can_execute_task(self):
        """
        Raises:
            (Exeption): If sample changer cannot execute a task
        """
        if not self.is_ready():
            raise Exception(
                "Cannot execute task: bad state ("
                + SampleChangerState.tostring(self.state)
                + ")"
            )

    def is_task_finished(self):
        """
        Returns:
            (str): Description of the error of last executed task (or None if success).
        """
        return self.is_ready() or (
            (not self.is_normal_state()) and (self.state != SampleChangerState.Unknown)
        )

    def is_executing_task(self):
        """
        Returns:
            (str): Description of the error of last executed task (or None if success).
        """
        return self.task is not None

    def wait_task_finished(self, timeout=None):
        """
        Wait for currently running task to finish.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting end of task")):
            while not self.is_task_finished():
                sleep(0.1)

    def get_loaded_sample(self):
        """
        Returns:
            (Sample) Currently loaded sample
        """
        for smp in self.get_sample_list():
            if smp.is_loaded():
                return smp
        return None

    def has_loaded_sample(self):
        """
        Returns:
         (boolean): True if a sample is loaded False otherwise
        """
        return self.get_loaded_sample() is not None

    def is_mounted_sample(self, sample_location):
        """Check if the sample is mounted.
        Args:
            sample_location: Sample location to check.
        Returns:
            (bool): True if mounted.
        """
        try:
            return self.get_loaded_sample().get_coords() == sample_location
        except AttributeError:
            return False

    def abort(self):
        """
        Aborts current task and puts device in safe state
        """
        self._do_abort()
        if self.task_proc is not None:
            self.task_proc.join(1.0)
            if self.task_proc is not None:
                self.task_proc.kill(Exception("Task aborted"))
                self.task = None
                self.task_proc = None
                self.task_error = None

    def update_info(self):
        """
        Update sample changer sample information, currently loaded sample
        and emits infoChanged and loadedSampleChanged when loaded sample
        have changed
        """
        former_loaded = self.get_loaded_sample()
        self._do_update_info()
        if self._is_dirty():
            self._trigger_info_changed_event()

        loaded = self.get_loaded_sample()
        if loaded != former_loaded:
            if (
                (loaded is None)
                or (former_loaded is None)
                or (loaded.get_address() != former_loaded.get_address())
            ):
                self._trigger_loaded_sample_changed_event(loaded)

        self._reset_dirty()

    def is_transient(self):
        """???"""
        return self._transient

    def _set_transient(self, value):
        """???"""
        self._transient = value

    def get_token(self):
        """???"""
        return self._token

    def set_token(self, token):
        """???"""
        self._token = token

    def get_sample_properties(self):
        """
        Returns:
            (tuple): With sample properties defined in Sample
        """
        return ()

    # ########################    TASKS    #########################
    def change_mode(self, mode, wait=True):
        """Change the mode (SC specific, imply change of the State)
        Args:
            mode (int):
                Modes:
                Unknown   = 0
                Normal    = 1
                Charging  = 2
                Disabled  = 3
            wait (boolean): True to block until mode changed is completed,
                            False otherwise
        Rerturns:
            (Object): Value returned by _execute_task either a Task
                      or result of the operation.
        """
        if mode == SampleChangerMode.Unknown:
            return
        if mode == self.get_state():
            return
        if self.get_state() == SampleChangerState.Disabled:
            self._set_state(SampleChangerState.Unknown)
            self.update_info()
        elif mode == SampleChangerMode.Disabled:
            self._set_state(SampleChangerState.Disabled)
        return self._execute_task(
            SampleChangerState.ChangingMode, wait, self._do_change_mode, mode
        )

    @dtask
    def scan(self, component=None, recursive=False):
        """Scan component or list of components for presence.
        Args:
            component (Component): Root component to start scan from. Sample
                                   changer root is used if None is passed.
            (recursive) (boolean): Recurse down the component structure if True,
                                   scan only component otherwise.
        Rerturns:
            (Object): Value returned by _execute_task either a Task
                      or result of the operation.
        """
        obj_list = []
        if isinstance(component, list):
            for comp in component:
                obj_list.append(self._scan_one(comp, recursive))
            return obj_list
        return self._scan_one(component, recursive)

    def _scan_one(self, component, recursive):
        """Scan component or list of components for samples.
        Args:
            component (Component): Root component to start scan from. Sample
                                   changer root is used if None is passed
            (recursive) (boolean): Recurse down the component structure if True.
                                   scan only component  otherwise.
        Rerturns:
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        self.assert_not_charging()
        if component is None:
            component = self
        component = self._resolve_component(component)
        component.assert_is_scannable()
        return self._execute_task(
            SampleChangerState.Scanning, True, self._do_scan, component, recursive
        )

    def select(self, component, wait=True):
        """
        Select a component.

        Args:
            component (Component): Component to select
            wait (boolean): True to wait for selection to complete otherwise False

        Rerturns:
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        component = self._resolve_component(component)
        ret = self._execute_task(
            SampleChangerState.Selecting, wait, self._do_select, component
        )
        self._trigger_selection_changed_event()
        return ret

    def chained_load(self, sample_to_unload, sample_to_load):
        """
        Chain the unload of a sample with a load.

        Args:
            sample_to_unload (tuple): sample address on the form
                                      (component1, ... ,component_N-1, component_N)
            sample_to_load (tuple): sample address on the form
                                      (component1, ... ,component_N-1, component_N)
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        self.unload(sample_to_unload)
        self.wait_ready(timeout=10)
        return self.load(sample_to_load)

    def load(self, sample=None, wait=True):
        """
        Load a sample.

        Args:
            sample (tuple): sample address on the form
                            (component1, ... ,component_N-1, component_N)
            wait (boolean): True to wait for load to complete False otherwise

        Returns
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        sample = self._resolve_component(sample)
        self.assert_not_charging()
        # Do a chained load in this case
        if self.has_loaded_sample():
            # Do a chained load in this case
            if (sample is None) or (sample == self.get_loaded_sample()):
                raise Exception(
                    "The sample "
                    + str(self.get_loaded_sample().get_address())
                    + " is already loaded"
                )
            return self.chained_load(self.get_loaded_sample(), sample)
        return self._execute_task(
            SampleChangerState.Loading, wait, self._do_load, sample
        )

    def unload(self, sample_slot=None, wait=True):
        """
        Unload sample to location sample_slot, unloads to the same slot as it
        was loaded from if None is passed

        Args:
            sample_slot (tuple): sample address on the form
                               (component1, ... ,component_N-1, component_N)
            wait: If True wait for unload to finish otherwise return immediately

        Returns:
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        sample_slot = self._resolve_component(sample_slot)
        self.assert_not_charging()
        # In case we have manually mounted we can command an unmount
        if not self.has_loaded_sample():
            raise Exception("No sample is loaded")
        return self._execute_task(
            SampleChangerState.Unloading, wait, self._do_unload, sample_slot
        )

    def reset(self, wait=True):
        """
        Reset the sample changer.

        wait: If True wait for reset to finish otherwise return immediately

        Returns:
            (Object): Value returned by _execute_task either a Task or result of the
                      operation
        """
        return self._execute_task(SampleChangerState.Resetting, wait, self._do_reset)

    def _load(self, sample=None):
        self._do_load(sample)

    def _unload(self, sample_slot=None):
        self._do_unload(sample_slot)

    def _resolve_component(self, component):
        if component is not None and isinstance(component, str):
            comp = self.get_component_by_address(component)
            if comp is None:
                raise Exception(f"Invalid component: {component}")
            return comp
        return component

    # ########################    ABSTRACTS    #########################

    @abc.abstractmethod
    def _do_abort(self):
        """
        Aborts current task and puts device in safe state
        """
        return

    @abc.abstractmethod
    def _do_update_info(self):
        return

    @abc.abstractmethod
    def _do_change_mode(self, mode):
        return

    @abc.abstractmethod
    def _do_scan(self, component, recursive):
        return

    @abc.abstractmethod
    def _do_select(self, component):
        return

    @abc.abstractmethod
    def _do_load(self, sample):
        return

    @abc.abstractmethod
    def _do_unload(self, sample_slot=None):
        return

    @abc.abstractmethod
    def _do_reset(self):
        return

    # ########################    PROTECTED    #########################

    def _execute_task(self, task, wait, method, *args):
        self.assert_can_execute_task()
        msg = f"Start {SampleChangerState.tostring(task)}"
        logging.debug(msg)
        self.task = task
        self.task_error = None
        self._set_state(task)
        ret = self._run(task, method, wait=False, *args)
        self.task_proc = ret

        ret.link(self._on_task_ended)
        if wait:
            return ret.get()
        return ret

    @dtask
    def _run(self, task, method, *args):
        """
        method(self,*arguments)
        exeption=None
        try:
            while !_is_task_finished(state):
              time.sleep(0.1)
            exeption=_getTaskException(state)
        finally:
            _trigger_task_finished_event(state,exeption)
            self._set_state(SampleChangerState.Ready)
        """
        exception = None
        ret = None
        try:
            ret = method(*args)
        except Exception as ex:
            exception = ex
        # if self.get_state()==self.task:
        #    self._set_state(SampleChangerState.Ready)
        self.update_info()
        task = self.task
        self.task = None
        self.task_proc = None
        self._trigger_task_finished_event(task, ret, exception)
        if exception is not None:
            self._on_task_failed(task, exception)
            raise exception
        return ret

    def _on_task_failed(self, task, exception):
        """What to do when task failed"""

    def _on_task_ended(self, task):
        """What to do when task ended normally"""
        try:
            msg = f"Task ended. Return value: {task.get()}"
            logging.debug(msg)
        except Exception as err:
            msg = f"Error while executing sample changer task: {err}"
            logging.error(msg)

    def _set_state(self, state=None, status=None):
        """Set the state"""
        if (state is not None) and (self.state != state):
            former = self.state
            self.state = state
            if status is None:
                status = SampleChangerState.tostring(state)
            self._trigger_state_changed_event(former)

        if (status is not None) and (self.status != status):
            self.status = status
            self._trigger_status_changed_event()

    def _reset_loaded_sample(self):
        for smp in self.get_sample_list():
            smp._set_loaded(False)
        self._trigger_loaded_sample_changed_event(None)

    def _set_loaded_sample(self, sample):
        previous_loaded = None

        for smp in self.get_sample_list():
            if smp.is_loaded():
                previous_loaded = smp
                break

        for smp in self.get_sample_list():
            if smp != sample:
                smp._set_loaded(False)
            else:
                if self.get_loaded_sample() == smp:
                    smp._set_loaded(True)

        if previous_loaded != self.get_loaded_sample():
            self._trigger_loaded_sample_changed_event(sample)

    def _set_selected_sample(self, sample):
        cur = self.get_selected_sample()
        if cur != sample:
            Container._set_selected_sample(self, sample)
            self._trigger_selection_changed_event()

    def _set_selected_component(self, component):
        cur = self.get_selected_component()
        if cur != component:
            Container._set_selected_component(self, component)
            self._trigger_selection_changed_event()

    def trigger_progress_message(self, message: str):
        self.emit(self.PROGRESS_MESSAGE, (message,))

    # ########################    PRIVATE    #########################

    def _trigger_state_changed_event(self, former):
        self.emit(self.STATE_CHANGED_EVENT, (self.state, former))

    def _trigger_status_changed_event(self):
        self.emit(self.STATUS_CHANGED_EVENT, (str(self.status),))

    def _trigger_loaded_sample_changed_event(self, sample):
        self.emit(self.LOADED_SAMPLE_CHANGED_EVENT, (sample,))

    def _trigger_selection_changed_event(self):
        self.emit(self.SELECTION_CHANGED_EVENT, ())

    def _trigger_info_changed_event(self):
        self.emit(self.INFO_CHANGED_EVENT, ())

    def _trigger_task_finished_event(self, task, ret, exception):
        self.emit(self.TASK_FINISHED_EVENT, (task, ret, exception))

    def _trigger_contents_updated_event(self):
        self.emit(self.CONTENTS_UPDATED_EVENT)



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import ast
from collections import OrderedDict

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.NState import NState


class P11Collimator(NState):
    """Collimator hardware object class"""

    def init(self):
        """Initialize the collimator with its motors and positions."""
        super().init()

        # Retrieve motors using their roles
        self.y_motor = self.get_object_by_role("collimatory")  # Y-axis motor
        self.z_motor = self.get_object_by_role("collimatorz")  # Z-axis motor

        self.log.info(f"Collimator Y Motor initialized: {self.y_motor}")
        self.log.info(f"Collimator Z Motor initialized: {self.z_motor}")

        # Load positions from XML configuration
        self.load_positions()

        # Load deltas for each motor
        self.load_deltas()

        # Set _positions for UI access
        self._positions = OrderedDict()
        self._positions = self.positions

    def load_positions(self):
        """Load predefined positions from the XML configuration."""
        self.log.info("Loading collimator positions from config")
        positions_str = self.get_property("values")

        # Convert the string to a dictionary using ast.literal_eval
        try:
            self.positions = ast.literal_eval(positions_str)
            if isinstance(self.positions, dict):
                self.log.info(f"Available collimator positions: {self.positions}")
            else:
                raise ValueError("Positions data is not a dictionary")
        except (SyntaxError, ValueError) as e:
            self.log.error(f"Error parsing collimator positions: {e}")
            raise ValueError(
                "Invalid collimator positions format in the configuration."
            )

    def load_deltas(self):
        """Load individual motor deltas from the XML configuration explicitly."""
        self.log.info("Loading deltas from config")

        # Fetch individual deltas for each motor
        delta_y = self.get_property("delta_collimatory")
        delta_z = self.get_property("delta_collimatorz")

        # If a delta is not specified, fallback to a default delta value
        self.deltas = {
            "collimatory": (
                float(delta_y) if delta_y is not None else self.default_delta
            ),
            "collimatorz": (
                float(delta_z) if delta_z is not None else self.default_delta
            ),
        }

        # Log the deltas for each motor
        for motorname, delta in self.deltas.items():
            self.log.info(f"Delta for {motorname}: {delta}")

    def set_value(self, value):
        """Set the collimator to the specified position."""
        if value not in self.positions:
            raise ValueError(f"Invalid state value: {value}")

        y_position = self.positions[value]["collimatory"]
        z_position = self.positions[value]["collimatorz"]

        # Move the motors
        self.y_motor._set_value(y_position)
        self.z_motor._set_value(z_position)
        self.log.info(f"Setting collimator to position: {value}")

    def get_position_list(self):
        """Return the list of available collimator positions."""
        return list(self.positions.keys())

    def get_value(self):
        """Get the current collimator position based on the motor positions."""
        current_y = self.y_motor.get_value()
        current_z = self.z_motor.get_value()

        for position_name, position in self.positions.items():
            if self.is_within_deltas(
                position.get("collimatory"), current_y, "collimatory"
            ) and self.is_within_deltas(
                position.get("collimatorz"), current_z, "collimatorz"
            ):
                return position_name  # Return the matching position name

    def is_within_deltas(self, target_value, current_value, motor_name):
        """Check if the current motor position is within the delta tolerance for that specific motor."""
        delta = self.deltas.get(motor_name)
        if target_value is None or delta is None:
            return False
        return abs(current_value - target_value) <= delta

    def is_moving(self):
        """Return True if any motor is moving."""
        return self.y_motor.is_moving() or self.z_motor.is_moving()

    def get_state(self):
        """Determine the overall state of the collimator motor system."""
        if self.is_moving():
            return HardwareObjectState.BUSY
        else:
            return HardwareObjectState.READY



import logging
import time

from mxcubecore.HardwareObjects.LNLS.LNLSMotor import LNLSMotor


class LNLSDetDistMotor(LNLSMotor):

    PMAC_SENDCMD = 'pmac_sendcmd'

    def __init__(self, name):
        LNLSMotor.__init__(self, name)

    def _move(self, value):
        """Override super class method."""
        if not self.validate_value(value):
            raise ValueError("Invalid value %s; limits are %s"
                             % (value, self.get_limits())
                             )
        self.update_specific_state(self.SPECIFIC_STATES.MOVING)

        for i in range(2):
            logging.getLogger(
                "user_level_log").info('Setting detector distance...'
            )

        # Enable air
        logging.getLogger("HWR").info('%s: Enabling air' % self.motor_name)
        command = '#5,7,8j/'
        self.set_channel_value(self.PMAC_SENDCMD, command)
        time.sleep(2)

        # Wait for stability
        logging.getLogger("HWR").info('%s: Wait for air stability' % self.motor_name)
        while (self.get_channel_value(self.MOTOR_DMOV) == 0):
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Move det dist motor
        logging.getLogger("HWR").info('%s: Set motor to %s' % (self.motor_name, value))
        self.set_channel_value(self.ACTUATOR_VAL, value)

        # Wait for movement
        logging.getLogger("HWR").info('%s: Wait for movement to finish' % self.motor_name)
        while (self.get_channel_value(self.MOTOR_DMOV) == 0
        or abs(current_value - value) > 0.0005):
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Disable air
        time.sleep(1)
        logging.getLogger("HWR").info('%s: Disabling air' % self.motor_name)
        command = '#5,7,8dkill'
        self.set_channel_value(self.PMAC_SENDCMD, command)
        time.sleep(0.5)

        # Wait for stability
        logging.getLogger("HWR").info(
            '%s: Wait for air stability again' % self.motor_name
        )
        while (self.get_channel_value(self.MOTOR_DMOV) == 0):
            time.sleep(0.2)
            current_value = self.get_value()
            self.update_value(current_value)

        # Wait for stability
        logging.getLogger("HWR").info('%s: Movement done!' % self.motor_name)
        self.update_state(self.STATES.READY)
        for i in range(2):
            logging.getLogger(
                "user_level_log").info('Detector distance is set.'
            )
        return value

    def _set_value(self, value):
        """Override method."""
        # As this motor moves in a special way, we delegate its movement
        # sequence to _move.
        pass



#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import sys

from mxcubecore.HardwareObjects.EMBL import EMBLQueueEntry
from mxcubecore.HardwareObjects.Gphl import GphlQueueEntry
from mxcubecore.model import queue_model_objects

# Import all constants and BaseQueueEntry from base_queue_entry so that
# queue _entry can be imported and used as before.
from mxcubecore.queue_entry.base_queue_entry import *
from mxcubecore.queue_entry.characterisation import CharacterisationGroupQueueEntry
from mxcubecore.queue_entry.import_helper import ImportHelper

# These two queue entries, for the moment violates the convention above and
# should eventually be changed
from mxcubecore.queue_entry.xrf_spectrum import XrfSpectrumQueueEntry

__all__ = []
MODEL_QUEUE_ENTRY_MAPPINGS = {}


def get_queue_entry_from_task_name(task_name):
    """
    Converts a snake cased task name to a camel cased class name, stripping
    all the "_" (underscores) and adding the suffix QueueEntry.

    Args:
        task_name (str): The task name

    Returns:
        str: The class name

    """
    cls_name = task_name.title().replace("_", "") + "QueueEntry"
    return getattr(sys.modules[__name__], cls_name, None)


def import_queue_entries(site_name_list=()):
    """
    Imports queue entries; imports all the native queue entries first and
    then the queue entries in site_name

    NBNB: Sets the globals __all__ and MODEL_QUEUE_ENTRY_MAPPINGS

    Args:
        site_name_list (list): List with sites to import from

    Returns:
        None

    """
    global __all__, MODEL_QUEUE_ENTRY_MAPPINGS

    # Import all queue entries, using the convention that the class name
    # is the camel cased modulename with a QueueEntry suffix.
    _modules = ImportHelper.import_queue_entry_modules(__name__)

    for site_name in site_name_list:
        _modules = ImportHelper.import_queue_entry_modules(__name__, site_name)

    __all__ = [*_modules]

    MODEL_QUEUE_ENTRY_MAPPINGS = {
        queue_model_objects.DataCollection: _modules["DataCollectionQueueEntry"],
        queue_model_objects.Characterisation: CharacterisationGroupQueueEntry,
        queue_model_objects.EnergyScan: _modules["EnergyScanQueueEntry"],
        queue_model_objects.XRFSpectrum: XrfSpectrumQueueEntry,
        queue_model_objects.SampleCentring: _modules["SampleCentringQueueEntry"],
        queue_model_objects.OpticalCentring: _modules["OpticalCentringQueueEntry"],
        queue_model_objects.DelayTask: DelayQueueEntry,
        queue_model_objects.Sample: SampleQueueEntry,
        queue_model_objects.Basket: BasketQueueEntry,
        queue_model_objects.TaskGroup: TaskGroupQueueEntry,
        queue_model_objects.Workflow: _modules["GenericWorkflowQueueEntry"],
        queue_model_objects.XrayCentering: _modules["XrayCenteringQueueEntry"],
        queue_model_objects.XrayCentring2: _modules["XrayCentering2QueueEntry"],
        queue_model_objects.GphlWorkflow: GphlQueueEntry.GphlWorkflowQueueEntry,
        queue_model_objects.XrayImaging: EMBLQueueEntry.XrayImagingQueueEntry,
    }

    # NBNB This is added for the queue models using pydantic (task_data
    # attribute) objects to define the data, to be able to use
    # MODEL_QUEUE_ENTRY_MAPPINGS for creating queue entries given a model.
    for _m in _modules.values():
        if hasattr(_m, "QMO"):
            MODEL_QUEUE_ENTRY_MAPPINGS[_m.QMO] = _m



import logging
import os
import time
from contextlib import ExitStack
from uuid import uuid1

import gevent
from gevent import subprocess
from lima2.client import Detector
from lima2.client.smx.aggregation import create_virtual_dataset

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.CommandContainer import ConnectionError
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector
from mxcubecore.TaskUtils import task

_logger = logging.getLogger("HWR")

_logger_det = logging.getLogger("lima2.client.detector")
_logger_smx = logging.getLogger("lima2.client.smx")

_logger_det.setLevel(logging.DEBUG)
_logger_smx.setLevel(logging.DEBUG)

# Logger decorator
def logger(fn):
    def inner(*args, **kwargs):
        _logger.debug(f"Entering %s", fn.__name__)
        to_execute = fn(*args, **kwargs)
        _logger.debug(f"Exiting %s", fn.__name__)
        return to_execute

    return inner

def convert_state(state):
    """Convert detector state to MxCube HWR state"""
    # UNKNOWN = 0
    # WARNING = 1
    # BUSY = 2
    # READY = 3
    # FAULT = 4
    # OFF = 5
    if state == Detector.State.IDLE or state == Detector.State.PREPARED:
        s = HardwareObjectState.READY
    elif state == Detector.State.RUNNING:
        s = HardwareObjectState.BUSY
    else:
        s = HardwareObjectState.UNKNOWN
    return s

def create_directory(path, check=True):
    subprocess.run(
        "mkdir --parents {0} && chmod -R 755 {0}".format(path),
        shell=True,
        check=check,
    )


class Lima2Detector(AbstractDetector):
    def __init__(self, name):
        AbstractDetector.__init__(self, name)
        self.header = dict()
        self.start_angles = list()

    def init(self):
        AbstractDetector.init(self)

        lima_ctrl_device = self.get_property("lima_ctrl_device", "")
        # lima_recv_devices = ast.literal_eval(self.get_property("lima_recv_devices", ""))
        lima_recv_devices = self.get_property("lima_recv_devices", "").split(",")

        _logger.info(
            f"Initializing Lima2Detector: {lima_ctrl_device} {lima_recv_devices}"
        )

        if not lima_ctrl_device or not len(lima_recv_devices) >= 1:
            return

        try:
            self.__device = Detector(lima_ctrl_device, *lima_recv_devices)

            self.__acq_params = self.__device.acq_params
            self.__proc_params = self.__device.proc_params

            # Monitor device state
            def on_state_change(state):
                s = convert_state(state)
                _logger.info(f"State changed to {state} / {s}")
                self.update_state(s)
            self.__device.registerStateLogger(on_state_change)

            self.__proc_loop_task = None
        except ConnectionError:
            self.update_state(HardwareObjectState.FAULT)
            _logger.error("Could not connect to detector %s" % lima_device)
            self._emit_status()

    def has_shutterless(self):
        return True

    @logger
    def wait_idle(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not idle")):
            idle_states = [Detector.State.IDLE, Detector.State.FAULT]
            while self.__device.state not in idle_states:
                print(self.__device.state)
                gevent.sleep(1)

            proc = self.__proc_loop_task
            if proc:
                _logger.debug("waiting for previous processing to finish")
                gevent.wait([proc])

    @logger
    def wait_prepared(self, timeout=3500):
        with gevent.Timeout(timeout, RuntimeError("Detector not prepared")):
            while self.__device.state != Detector.State.PREPARED:
                gevent.sleep(1)

    @logger
    def last_image_saved(self):
        try:
            img = 0  # TODO
            return img
        except Exception:
            return 0

    def get_deadtime(self):
        return float(self.get_property("deadtime"))

    def find_next_pedestal_dir(self, data_root_path, subdir):
        _index = 1
        _indes_str = "%04d" % _index
        fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        while os.path.exists(fpath):
            _index += 1
            _indes_str = "%04d" % _index
            fpath = os.path.join(data_root_path, f"{subdir}_{_indes_str}")

        return fpath

    def set_detector_filenames(self, data_root_path, prefix):
        create_directory(data_root_path)

    @logger
    def prepare_acquisition(self, number_of_images, exptime, data_root_path, prefix):
        packet_fifo_depth = 20000

        acq_params = {
            "acq": {
                "expo_time": int(exptime * 1e6),
                # "latency_time": 990,
                "nb_frames": number_of_images,
                "trigger_mode": "external",
                "nb_frames_per_trigger": 1,
            },
            "det": {
                "gain_mode": "dynamic",
                "packet_fifo_depth": packet_fifo_depth,
            },
        }

        self.set_detector_filenames(data_root_path, prefix)

        saving_groups = ["raw", "dense", "sparse", "accumulated"]

        # all streams will be saved in dedicated sub dirs except dense
        data_sub_dir = {g: g if g != "dense" else "" for g in saving_groups}

        data_path = {
            g: os.path.join(data_root_path, s) if s else data_root_path
            for g, s in data_sub_dir.items()
        }

        acc_frames = 1000

        max_nb_frames_per_file = dict(
            raw=1000,
            dense=1000,
            sparse=10000,
            accumulated=10,
        )

        nb_recvs = 2
        nb_recv_frames = number_of_images // nb_recvs
        nb_acc_frames = (number_of_images - 1) // acc_frames + 1

        nb_frames = {
            g: nb_acc_frames if g == "accumulated" else nb_recv_frames
            for g in saving_groups
        }

        def calc_frame_per_file(group, n):
            return min(max_nb_frames_per_file[group], n)

        frames_per_file = {
            g: calc_frame_per_file(g, nb_frames[g]) for g in saving_groups
        }

        save_files = dict(
            raw=False,
            dense=True,
            sparse=True,
            accumulated=True,
        )

        dense_comp_with_hw_nx = True
        dense_comp = "zip" if dense_comp_with_hw_nx else "bshuf_lz4"

        compression = dict(
            raw=dense_comp, dense=dense_comp, sparse="none", accumulated="zip"
        )

        saving_streams = ["raw", "dense", "sparse"] + [
            f"accumulation_{a}" for a in ["corrected", "peak"]
        ]

        def get_stream_group(stream):
            is_acc = stream.startswith("accumulation_")
            return "accumulated" if is_acc else stream

        def get_stream_prefix(stream):
            return stream.replace("accumulation", "acc")

        def get_saving(stream):
            group = get_stream_group(stream)
            stream_prefix = get_stream_prefix(stream)
            return dict(
                enabled=save_files[group],
                base_path=data_path[group],
                filename_prefix=f"{prefix}_{stream_prefix}",
                start_number=0,
                nb_frames_per_file=frames_per_file[group],
                file_exists_policy="abort",
                compression=compression[group],
            )

        sub_dirs = [s for g, s in data_sub_dir.items() if s and save_files[g]]
        # the sub directory for aggregated data from all receivers
        sub_dirs.append("aggregated")
        for s in sub_dirs:
            create_directory(os.path.join(data_root_path, s))

        fai_kernels_base = self.get_property("fai_kernels_base", "")
        params_base = self.get_property("params_base", "")
        pedestal_path = os.path.join(data_root_path, "pedestal.h5")

        mask_beam_stop = True
        mask_filename = "mask.h5" if mask_beam_stop else "mask_no_beamstop.h5"

        manage_proc = True

        energy = 11.56  # 9.06

        fai_params = {
            "mask_path": params_base + mask_filename,
            "csr_path": params_base + "csr.h5",
            "radius2d_path": params_base + "r_center.h5",
            "radius1d_path": params_base + "bin_centers.h5",
            "error_model": "azimuthal",
            "dummy": 0.0,
            "delta_dummy": 0.0,
            "normalization_factor": 1.0,
            "cutoff_clip": 5.0,
            "cycle": 5,
            # "empty": -9999.0,
            "noise": 1.0,
            "cutoff_pick": 3.0,
            "acc_nb_frames_reset": 0,
            "acc_nb_frames_xfer": acc_frames,
        }

        def get_dense_out_params(variant_name):
            legacy_photon_adus = 41.401 * energy

            variant_data = {
                "legacy": dict(
                    pixel_type="int32", photon_adus=legacy_photon_adus, photon_bias=0.0
                ),
                "0": dict(pixel_type="int16", photon_adus=16.0, photon_bias=0.0),
                "1": dict(pixel_type="uint16", photon_adus=16.0, photon_bias=32.0),
                "2": dict(pixel_type="uint16", photon_adus=8.0, photon_bias=32.0),
                "3": dict(pixel_type="int16", photon_adus=8.0, photon_bias=0.0),
            }

            variant = variant_data[variant_name]
            photon_adus = variant["photon_adus"]
            return dict(
                dense_intensity_factor=photon_adus,
                dense_intensity_offset=photon_adus * variant["photon_bias"],
                dense_pixel_type="dense_%s" % variant["pixel_type"],
            )

        dense_variant_name = "legacy"
        fai_params.update(get_dense_out_params(dense_variant_name))

        fai_params_fname = os.path.join(params_base, "fai_params.json")
        if os.path.exists(fai_params_fname):
            with open(fai_params_fname, "rt") as f:
                fai_params.update(json.loads(f.read()))

        proc_params = {
            "proc_mode": "fai",
            "fifo": {
                "nb_fifo_frames": packet_fifo_depth,
            },
            "buffers": {
                "nb_peak_counters_buffer": nb_recv_frames,
            },
            "gpu": {
                "device_idx": 0,
                "cl_source_path": fai_kernels_base,
            },
            "jfrau": {
                "gain_path": params_base + "gains.h5",
                "pedestal_path": pedestal_path,
                "photon_energy": energy,
            },
            "fai": fai_params,
        }

        proc_params.update(
            {f"saving_{stream}": get_saving(stream) for stream in saving_streams}
        )

        self.__device.nb_gpus_per_system = 2

        self.__device.acq_params.update(acq_params)
        _logger.debug("acq_params: %s", self.__device.acq_params)
        self.__device.proc_params.update(proc_params)
        _logger.debug("proc_params: %s", self.__device.proc_params)

        uuid = uuid1()
        _logger.info(f"UUID={uuid}")

        self.__device.prepareAcq(uuid)

        # Async version
        # gevent.spawn(self.__device.prepareAcq, uuid).link_value(on_prepared)

        # Manage processing
        if manage_proc:
            self.__proc_loop_task = gevent.spawn(self.proc_loop, uuid)

    @logger
    def start_acquisition(self):
        self.wait_prepared()

        self.__device.startAcq()

    @logger
    def stop_acquisition(self):
        wait_for_idle_timeout = 10
        try:
            self.wait_idle(wait_for_idle_timeout)
        except:
            _logger.warning("detector not idle after %s", wait_for_idle_timeout)
            pass

        try:
            self.__device.stopAcq()
        except Exception:
            if self.__proc_loop_task:
                gevent.kill(self.__proc_loop_task)
            self.__device.resetAcq()
        finally:
            self.wait_idle()

    @logger
    def proc_loop(self, uuid):
        with ExitStack() as stack:
            def cleanup():
                self.__proc_loop_task = None
            stack.callback(cleanup)

            _logger.debug("starting processing loop for %s", uuid)
            proc_cm = self.__device.getProcessing(uuid, erase_on_cleanup=True)
            proc = stack.enter_context(proc_cm)
            stack.callback(_logger.debug, "erasing processing for %s", uuid)

            # main processing loop
            while not proc.is_finished:
                gevent.sleep(1.0)

    @logger
    def reset(self):
        self.__device.resetAcq()

    @property
    def status(self):
        try:
            acq_status = self.__device.state
        except Exception:
            acq_status = "OFFLINE"

        status = {
            "acq_satus": str(acq_status).upper(),
        }

        return status

    def _emit_status(self):
        self.emit("statusChanged", self.status)



"""
Class for reading images from Falcon camera OAV
"""

from mxcubecore import BaseHardwareObjects
from mxcubecore import HardwareRepository as HWR


class XalocCalibration(BaseHardwareObjects.HardwareObject):
    def __init__(self, name):
        super().__init__(name)

    def init(self):

        self.calibx = self.get_channel_object("calibx")
        self.caliby = self.get_channel_object("caliby")

        if self.calibx is not None and self.caliby is not None:
            print("Connected to calibration channels")

    def getCalibration(self):
        return [self.calibx.get_value(), self.caliby.get_value()]


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    calib = hwr.get_hardware_object("/calibration")
    print("Calibration is: ", calib.getCalibration())


if __name__ == "__main__":
    test()



import logging

import gevent
from pydantic.v1 import (
    BaseModel,
    Field,
)
from typing_extensions import Literal

from mxcubecore.HardwareObjects.BeamlineActions import (
    AnnotatedCommand,
    BeamlineActions,
)


class SimpleFloat(BaseModel):
    exp_time: float = Field(100e-6, gt=0, lt=10, description="(s)")


class StringLiteral(BaseModel):
    phase: Literal["Centring", "DataCollection", "BeamLocation", "Transfer"] = Field(1)


class SimulatedAction:
    def __call__(self, *args, **kw):
        gevent.sleep(3)
        return args


class SimulatedActionError:
    def __call__(self, *args, **kw):
        raise RuntimeError("Simulated error")


class LongSimulatedAction:
    def __call__(self, *args, **kw):
        for i in range(10):
            gevent.sleep(1)
            logging.getLogger("user_level_log").info("%d, sleeping for 1 second", i + 1)

        return args


class Anneal2(AnnotatedCommand):
    def __init__(self, *args):
        super().__init__(*args)

    def anneal2(self, data: SimpleFloat) -> None:
        logging.getLogger("user_level_log").info(
            f"Annealing for {data.exp_time} seconds"
        )
        gevent.sleep(data.exp_time)


class QuickRealign2(AnnotatedCommand):
    def __init__(self, *args):
        super().__init__(*args)

    def quick_realign2(self) -> None:
        for i in range(10):
            gevent.sleep(1)
            logging.getLogger("user_level_log").info("%d, sleeping for 1 second", i + 1)


class ComboTest2(AnnotatedCommand):
    def __init__(self, *args):
        super().__init__(*args)

    def combo_test2(self, data: StringLiteral) -> None:
        logging.getLogger("user_level_log").info(f"Selected {data.phase}")


class BeamlineActionsMockup(BeamlineActions):
    def __init__(self, *args):
        super().__init__(*args)



"""Sample Changer Hardware Object
"""

import logging

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject


class GrobSampleChanger(HardwareObject):
    (FLAG_SC_IN_USE, FLAG_MINIDIFF_CAN_MOVE, FLAG_SC_CAN_LOAD, FLAG_SC_NEVER) = (
        1,
        2,
        4,
        8,
    )
    (STATE_BASKET_NOTPRESENT, STATE_BASKET_UNKNOWN, STATE_BASKET_PRESENT) = (-1, 0, 1)

    USE_SPEC_LOADED_SAMPLE = False
    ALWAYS_ALLOW_MOUNTING = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        self._procedure = ""
        self._successCallback = None
        self._failureCallback = None
        self._holderlength = 22
        self._sample_id = None
        self._sample_location = (0, 0)
        self.loaded_sample_dict = {}
        self.samples_map = dict(zip(range(30), ["unknown"] * 30))
        self.matrix_codes = []
        for i in range(30):
            sample_num = i + 1
            basket = int((sample_num - 1) / 10) + 1
            vial = 1 + ((sample_num - 1) % 10)
            self.matrix_codes.append(
                ("HA%d" % sample_num, basket, vial, "A%d" % basket, 0)
            )

        grob = self.get_object_by_role("grob")
        self.grob = grob.controller
        self.connect(self.grob, "transfer_state", self.sample_changer_state_changed)
        self.connect(self.grob, "io_bits", self.io_bits_changed)
        self.connect(self.grob, "mounted_sample", self.mounted_sample_changed)
        self.connect(self.grob, "samples_map", self.samples_map_changed)

    def connect_notify(self, signal):
        logging.info("%s: connect_notify %s", self.name(), signal)
        if signal == "stateChanged":
            self.sample_changer_state_changed(self.get_state())
        elif signal == "loadedSampleChanged":
            self.mounted_sample_changed(self._get_loaded_sampleNum())
        elif signal == "samples_map_changed":
            self.samples_map_changed(self.get_samples_map())

    def get_state(self):
        return self.grob.transfer_state()

    def io_bits_changed(self, bits):
        bits, output_bits = map(int, bits.split())
        status = {}
        for bit_number, name in {
            1: "lid",
            2: "puck1",
            3: "puck2",
            4: "puck3",
            6: "ln2_alarm_low",
        }.items():
            status[name] = bits & (1 << (bit_number - 1)) != 0
        self.emit("ioStatusChanged", (status,))

    def samples_map_changed(self, samples_map_dict):
        samples_map_int_keys = {}
        for k, v in samples_map_dict.items():
            samples_map_int_keys[int(k)] = v
        self.samples_map = samples_map_int_keys
        self.emit("samples_map_changed", (self.samples_map,))

    def get_samples_map(self):
        return self.samples_map

    def mounted_sample_changed(self, sample_num=None):
        self.emit("sampleIsLoaded", (sample_num > 0,))

        if sample_num > 0:
            basket = int((sample_num - 1) / 10) + 1
            vial = 1 + ((sample_num - 1) % 10)

    def sample_changer_state_changed(self, state):
        self.emit("stateChanged", (state,))

    def _callSuccessCallback(self):
        if callable(self._successCallback):
            try:
                self._successCallback()
            except Exception:
                logging.exception(
                    "%s: exception while calling success callback", self.name()
                )

    def _call_failure_callback(self):
        if callable(self._failureCallback):
            try:
                self._failureCallback()
            except Exception:
                logging.exception(
                    "%s: exception while calling failure callback", self.name()
                )

    def _sample_transfer_done(self, transfer_greenlet):
        status = transfer_greenlet.get()
        if status == "READY":
            self.prepare_centring()
            self.sample_changer_state_changed("READY")
            self._callSuccessCallback()
        else:
            self.sample_changer_state_changed("ERROR")
            self._call_failure_callback()

    def prepare_centring(self):
        pass

    def get_loaded_sample(self):
        return self.loaded_sample_dict

    def _set_moving_state(self):
        self.sample_changer_state_changed("MOVING")

    def _get_loaded_sampleNum(self):
        samples_map = self.get_samples_map()
        for i in range(30):
            if samples_map[i] == "on_axis":
                return i + 1

    def unload_mounted_sample(
        self,
        holderLength=None,
        sample_id=None,
        sample_location=None,
        sampleIsUnloadedCallback=None,
        failureCallback=None,
    ):
        self._procedure = "UNLOAD"

        self._successCallback = sampleIsUnloadedCallback
        self._failureCallback = failureCallback

        gevent.spawn(self.continue_transfer, self.prepare_transfer()).link(
            self._sample_transfer_done
        )

    def load(
        self,
        sample=None,
        sample_id=None,
        holderLength=None,
        successCallback=None,
        failureCallback=None,
        prepareCentring=None,
        prepareCentringMotors={},
        prepare_centring=None,
        prepare_centring_motors=None,
        wait=True,
    ):
        self._successCallback = successCallback
        self._failureCallback = failureCallback
        self._holderlength = holderLength
        self._sample_id = sample_id
        self._sample_location = sample

        if self._get_loaded_sampleNum():
            self._procedure = "UNLOAD_LOAD"
        else:
            self._procedure = "LOAD"

        gevent.spawn(self.continue_transfer, self.prepare_transfer()).link(
            self._sample_transfer_done
        )

    def prepare_transfer(self):
        return True

    def continue_transfer(self, ok):
        if not ok:
            self.sample_changer_state_changed("ERROR")
            self._call_failure_callback()
            return

        basket, vial = self._sample_location
        sample_num = (basket - 1) * 10 + vial

        if self._procedure == "LOAD":
            logging.info("asking robot to load sample %d", sample_num)
            return self.grob.mount(sample_num)
        elif self._procedure == "UNLOAD_LOAD":
            sample_to_unload_num = self._get_loaded_sampleNum()
            logging.info(
                "asking robot to unload sample %d and to load sample %d",
                sample_to_unload_num,
                sample_num,
            )
            self.grob.unmount(sample_to_unload_num)
            return self.grob.mount(sample_num)
        elif self._procedure == "UNLOAD":
            sample_num = self._get_loaded_sampleNum()
            logging.info("asking robot to unload sample %d", sample_num)
            return self.grob.unmount(sample_num)

    def is_microdiff(self):
        return False

    def get_loaded_sampleDataMatrix(self):
        return None

    def get_loaded_sampleLocation(self):
        sample_num = self._get_loaded_sampleNum()
        if sample_num < 0:
            return None
        basket = int((sample_num - 1) / 10) + 1
        vial = 1 + ((sample_num - 1) % 10)
        return (basket, vial)

    def get_loaded_holder_length(self):
        return self._holderlength

    def get_matrix_codes(self):
        return self.matrix_codes

    def update_data_matrices(self):
        self.emit("matrixCodesUpdate", (self.matrix_codes,))

    def canLoadSample(self, sample_code=None, sample_location=None, holder_length=None):
        already_loaded = False

        loaded_sample_location = self.get_loaded_sampleLocation()

        if loaded_sample_location is not None:
            if (
                sample_location is not None
                and sample_location == loaded_sample_location
            ):
                already_loaded = True

        return (True, already_loaded)

    def open_dewar(self, callback):
        gevent.spawn(self.grob.open_dewar).link(callback)

    def close_dewar(self, callback):
        gevent.spawn(self.grob.close_dewar).link(callback)



import logging
import time

from mxcubecore.BaseHardwareObjects import HardwareObject

"""
Use the exporter to set different MD2 actuators in/out.
If private_state not specified, True will be send to set in and False for out.
Example xml file:
<object class="MicrodiffInOut">
  <username>Scintilator</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <cmd_name>ScintillatorPosition</cmd_name>
  <private_state>{"PARK":"out", "SCINTILLATOR":"in"}</private_state>
  <use_hwstate>True</use_hwstate>
</object>
"""


class MicrodiffInOutMockup(HardwareObject):
    def __init__(self, name):
        super().__init__(name)
        self.actuatorState = "unknown"
        self.username = "unknown"
        # default timeout - 3 sec
        self.timeout = 3
        self.hwstate_attr = None

    def init(self):
        self.cmd_name = self.get_property("cmd_name")
        self.username = self.get_property("username")

        self.states = {True: "in", False: "out"}
        self.state_attr = False
        self.offset = self.get_property("offset", 0)
        if self.offset > 0:
            self.states = {self.offset: "out", self.offset - 1: "in"}

        states = self.get_property("private_state")
        if states:
            import ast

            self.states = ast.literal_eval(states)
        try:
            tt = float(self.get_property("timeout"))
            self.timeout = tt
        except Exception:
            pass

        self.moves = dict((self.states[k], k) for k in self.states)

    def connect_notify(self, signal):
        if signal == "actuatorStateChanged":
            self.value_changed(self.state_attr)

    def value_changed(self, value):
        self.actuatorState = self.states.get(value, "unknown")
        self.emit("actuatorStateChanged", (self.actuatorState,))

    def _ready(self):
        return True

    def _wait_ready(self, timeout=None):
        timeout = timeout or self.timeout
        tt1 = time.time()
        while time.time() - tt1 < timeout:
            if self._ready():
                break
            else:
                time.sleep(0.5)

    def get_actuator_state(self, read=False):
        if read is True:
            value = self.state_attr
            self.actuatorState = self.states.get(value, "unknown")
            self.connect_notify("actuatorStateChanged")
        else:
            if self.actuatorState == "unknown":
                self.connect_notify("actuatorStateChanged")
        return self.actuatorState

    def actuatorIn(self, wait=True, timeout=None):
        if self._ready():
            try:
                self.state_attr = self.moves["in"]
                if wait:
                    timeout = timeout or self.timeout
                    self._wait_ready(timeout)
                self.value_changed(self.state_attr)
            except Exception:
                logging.getLogger("user_level_log").error(
                    "Cannot put %s in", self.username
                )
        else:
            logging.getLogger("user_level_log").error(
                "Microdiff is not ready, will not put %s in", self.username
            )

    def actuatorOut(self, wait=True, timeout=None):
        if self._ready():
            try:
                self.state_attr = self.moves["out"]
                if wait:
                    timeout = timeout or self.timeout
                    self._wait_ready(timeout)
                self.value_changed(self.state_attr)
            except Exception:
                logging.getLogger("user_level_log").error(
                    "Cannot put %s out", self.username
                )
        else:
            logging.getLogger("user_level_log").error(
                "Microdiff is not ready, will not put %s out", self.username
            )



import logging

from energy import energy
from scipy.constants import (
    angstrom,
    c,
    eV,
    h,
    kilo,
)

from mxcubecore.HardwareObjects.mockup.EnergyMockup import EnergyMockup


class PX2Energy(EnergyMockup):
    def init(self):
        self.energy = energy()

        self.energy_channel = self.get_channel_object("energy")
        self.energy_channel.connect_signal("update", self.energy_changed)

        self.state_channel = self.get_channel_object("state")
        self.state_channel.connect_signal("update", self.energy_state_changed)

        self.tunable = self.get_property("tunable")
        self.default_en = self.get_property("default_energy")

        self.minimum_energy = self.get_property("min_energy")
        self.maximum_energy = self.get_property("max_energy")

        self.current_energy = self.energy.get_value() / kilo
        self.current_wavelength = self.get_wavelegth_from_energy(self.current_energy)

        self.checkLimits = self.check_limits
        self.cancelMoveEnergy = self.cancel_move_energy

    def re_emit_values(self):
        self.emit("energyChanged", (self.current_energy, self.current_wavelength))
        self.emit("valueChanged", (self.current_energy,))

    def abort(self):
        self.energy.abort()

    def get_current_energy(self):
        return self.energy.get_energy() / kilo

    def get_wavelegth_from_energy(self, energy):
        return (h * c) / (eV * angstrom * kilo) / energy

    def get_energy_from_wavelength(self, wavelength):
        return (h * c) / (eV * angstrom * kilo) / wavelength

    def get_wavelength(self):
        return self.get_wavelegth_from_energy(self.get_current_energy())

    def get_limits(self):
        return self.minimum_energy, self.maximum_energy

    def get_wavelength_limits(self):
        minimum_wavelength = self.get_wavelegth_from_energy(self.maximum_energy)
        maximum_wavelength = self.get_wavelegth_from_energy(self.minimum_energy)
        return minimum_wavelength, maximum_wavelength

    def set_value(self, energy):
        logging.getLogger("user_level_log").info("Move energy to %6.3f keV" % energy)
        self.emit("moveEnergyStarted", ())
        self.energy.set_energy(energy)
        self.current_energy = self.get_current_energy()
        self.current_wavelength = self.get_wavelegth_from_energy(self.current_energy)
        self.re_emit_values()
        self.emit("moveEnergyFinished", ())

    def set_wavelength(self, wavelength):
        energy = self.get_energy_from_wavelength(wavelength)
        self.set_value(energy)

    def check_limits(self, value):
        logging.getLogger("HWR").debug("Checking the move limits")
        en_lims = self.get_limits()
        if value >= self.en_lims[0] and value <= self.en_lims[1]:
            logging.getLogger("HWR").info("Limits ok")
            return True
        logging.getLogger("user_level_log").info("Requested value is out of limits")
        return False

    def cancel_move_energy(self):
        logging.getLogger("user_level_log").info("Cancel energy move")
        self.abort()

    def energy_changed(self, pos):
        # logging.getLogger('HWR').info("energy_changed %s" % str(pos))
        energy = pos
        try:
            if abs(energy - self.current_energy) > 1e-3:
                self.current_energy = energy
                self.current_wavelength = 12.3984 / energy
                if self.current_wavelength is not None:
                    self.re_emit_values()
        except Exception:
            logging.getLogger("HWR").info(
                "energy_changed: error occured during an energy update"
            )

    def energy_state_changed(self, state):
        logging.getLogger("HWR").info("energy_state_changed %s" % str(state))
        # self.energy_server_check_for_errors(state)
        if state == "STANDBY":
            if self.moving:
                self.moving = False
                self.set_break_bragg()
            self.move_energy_finished(0)
            self.emit("stateChanged", "ready")
            self.emit("statusInfoChanged", "")
        elif state in ["MOVING", "ALARM"]:
            self.move_energy_started()
            self.emit("stateChanged", "busy")
